import os
import subprocess
import logging
import sys
import shutil

sys.path.insert(1, "/home/coder/workspace/white-box-testing/testframeworkwhiteboxmodul")
from unitTestGenerator.unit_test_generator import UnitTestGenerator
from config import Config

def read_code_samples(projects, defects4j_home):
    """
    Read Java code samples from Defects4J projects.
    """
    file_contents = []
    project_versions = {
        'Lang': 1,
        'Chart': 1,
        'Cli': 1,
        'Csv': 1,
        'Gson': 1
    }
    working_dir_base = os.path.join(defects4j_home, 'working_dir')
    os.makedirs(working_dir_base, exist_ok=True)

    for project in projects:
        version = project_versions.get(project, 1)
        checkout_dir = os.path.join(working_dir_base, f'{project}_{version}')
        if os.path.exists(checkout_dir):
            logging.info(f"Project {project} version {version} already checked out.")
        else:
            # Checkout the project
            logging.info(f"Checking out {project} version {version}...")
            try:
                subprocess.run(
                    ['defects4j', 'checkout', '-p', project, '-v', f'{version}b', '-w', checkout_dir],
                    check=True
                )
                logging.info(f"Checked out {project} version {version}.")
            except subprocess.CalledProcessError as e:
                logging.error(f"Failed to checkout {project} version {version}: {e}")
                continue  # Skip to the next project

        # Potential source directories
        possible_src_dirs = [
            os.path.join(checkout_dir, 'src', 'java'),
            os.path.join(checkout_dir, 'src', 'main', 'java'),
            os.path.join(checkout_dir, 'src')
        ]
        src_dir = None
        for dir_path in possible_src_dirs:
            if os.path.exists(dir_path):
                src_dir = dir_path
                break

        if src_dir is None:
            logging.warning(f"No source directory found for project {project} version {version}.")
            continue

        for root, _, files in os.walk(src_dir):
            for file in files:
                if file.endswith('.java'):
                    class_file_path = os.path.join(root, file)
                    try:
                        with open(class_file_path, 'r') as f:
                            content = f.read()
                            # Get the fully qualified class name
                            package = get_package_name(content)
                            class_name = file[:-5]  # Remove '.java'
                            if package:
                                full_class_name = f'{package}.{class_name}'
                            else:
                                full_class_name = class_name
                            file_contents.append((project, full_class_name, content))
                            logging.debug(f"Added class {full_class_name} from project {project}.")
                    except Exception as e:
                        logging.error(f"Error reading file '{class_file_path}': {e}")

    logging.info(f"Total classes collected: {len(file_contents)}")
    return file_contents

def setup_defects4j_env(defects4j_home: str) -> dict:
    """Setup Defects4J environment variables."""
    env_vars = os.environ.copy()
    env_vars['DEFECTS4J_HOME'] = defects4j_home
    env_vars['TZ'] = 'America/Los_Angeles'
    
    # Add defects4j binary to PATH if not already there
    defects4j_bin = os.path.join(defects4j_home, 'framework', 'bin')
    if defects4j_bin not in env_vars.get('PATH', ''):
        env_vars['PATH'] = f"{defects4j_bin}:{env_vars.get('PATH', '')}"
    
    # Debug information
    logging.debug(f"Current PATH: {env_vars.get('PATH', '')}")
    logging.debug(f"Current DEFECTS4J_HOME: {env_vars.get('DEFECTS4J_HOME', '')}")
    
    return env_vars

def check_defects4j_installation():
    """Check if defects4j is properly installed and accessible."""
    try:
        # Check if DEFECTS4J_HOME is in environment
        defects4j_home = os.environ.get('DEFECTS4J_HOME')
        if not defects4j_home:
            logging.error("DEFECTS4J_HOME environment variable is not set in environment")
            return False
            
        logging.info(f"Using DEFECTS4J_HOME: {defects4j_home}")
        
        # Check if the directory exists
        if not os.path.exists(defects4j_home):
            logging.error(f"DEFECTS4J_HOME directory does not exist: {defects4j_home}")
            return False
            
        # Check if defects4j binary is in path
        defects4j_bin = os.path.join(defects4j_home, 'framework', 'bin', 'defects4j')
        if not os.path.exists(defects4j_bin):
            logging.error(f"defects4j binary not found at: {defects4j_bin}")
            return False
            
        # Try running defects4j command
        result = subprocess.run(
            [defects4j_bin, 'env'], 
            check=True, 
            capture_output=True,
            text=True,
            env=os.environ
        )
        logging.info(f"Defects4J binary found and executable")
        return True
        
    except (subprocess.CalledProcessError, FileNotFoundError) as e:
        logging.error(f"Error checking Defects4J installation: {str(e)}")
        return False

def clean_working_directory(checkout_dir):
    """Clean up working directory if it exists."""
    if os.path.exists(checkout_dir):
        try:
            shutil.rmtree(checkout_dir)
            logging.info(f"Cleaned up existing directory: {checkout_dir}")
        except Exception as e:
            logging.error(f"Failed to clean up directory {checkout_dir}: {e}")

def generate_unit_tests(defects4j_home):
    """Generate unit tests for the 'Chart' project using UnitTestGenerator."""
    # Setup environment variables first
    new_env = setup_defects4j_env(defects4j_home)
    os.environ.clear()
    os.environ.update(new_env)
    
    if not check_defects4j_installation():
        return

    project = 'Chart'
    version = '1'
    working_dir_base = os.path.join(defects4j_home, 'working_dir')
    checkout_dir = os.path.join(working_dir_base, f'{project}_{version}')

    # Clean up existing checkout directory
    clean_working_directory(checkout_dir)
    os.makedirs(checkout_dir, exist_ok=True)

    # Set up environment variables
    env = {
        'TZ': 'America/Los_Angeles',
        'DEFECTS4J_HOME': defects4j_home,
        **os.environ
    }

    # Checkout the project
    try:
        logging.info(f"Checking out {project} version {version}...")
        checkout_result = subprocess.run(
            ['defects4j', 'checkout', '-p', project, '-v', f'{version}b', '-w', checkout_dir],
            check=True,
            env=env,
            capture_output=True,
            text=True
        )
        logging.info(f"Checkout output: {checkout_result.stdout}")
    except subprocess.CalledProcessError as e:
        logging.error(f"Failed to checkout: {e.stdout}\n{e.stderr}")
        return

    # Compile the project
    try:
        compile_result = subprocess.run(
            ['defects4j', 'compile'],
            cwd=checkout_dir,
            check=True,
            env=env,
            capture_output=True,
            text=True
        )
        logging.info(f"Compilation output: {compile_result.stdout}")
    except subprocess.CalledProcessError as e:
        logging.error(f"Failed to compile: {e.stdout}\n{e.stderr}")
        return

    # Get source directory
    try:
        export_result = subprocess.run(
            ['defects4j', 'export', '-p', 'dir.src.classes'],
            cwd=checkout_dir,
            check=True,
            env=env,
            capture_output=True,
            text=True
        )
        src_relative_path = export_result.stdout.strip()
        src_dir = os.path.join(checkout_dir, src_relative_path)
        
        if not os.path.exists(src_dir):
            logging.error(f"Source directory does not exist: {src_dir}")
            return
            
        logging.info(f"Using source directory: {src_dir}")
    except subprocess.CalledProcessError as e:
        logging.error(f"Failed to get source directory: {e.stdout}\n{e.stderr}")
        return

    # Initialize and run the unit test generator
    logging.info(f"Generating unit tests for project {project}...")
    config = Config()
    storage_path = os.path.join(defects4j_home, 'generated_tests')
    os.makedirs(storage_path, exist_ok=True)
    unit_test_generator = UnitTestGenerator(config=config, storage_path=storage_path)
    result = unit_test_generator.run(project_path=src_dir)
    
    if result.get('status') == 'success':
        logging.info(f"Unit tests generated for project {project}.")

        # Save the generated tests into the source folder
        generated_tests = result.get('generated_tests', [])
        for idx, (language, test_code, method_data) in enumerate(generated_tests, start=1):
            test_class_name = f"{method_data.get_class_name()}Test"
            test_file_name = f"{test_class_name}.java"
            test_file_path = os.path.join(src_dir, test_file_name)
            try:
                with open(test_file_path, 'w') as test_file:
                    test_file.write(test_code)
                logging.info(f"Saved test file: {test_file_path}")
            except Exception as e:
                logging.error(f"Failed to save test file {test_file_name}: {e}")
    else:
        logging.error(f"Failed to generate unit tests for project {project}: {result.get('error')}")

if __name__ == "__main__":
    logging.basicConfig(level=logging.DEBUG)  # Changed to DEBUG level
    defects4j_home = "/home/coder/workspace/defects4j"
    
    # Verify defects4j home exists
    if not os.path.exists(defects4j_home):
        logging.error(f"Defects4J home directory not found: {defects4j_home}")
        logging.error("Please make sure Defects4J is properly installed")
        sys.exit(1)
    
    # Print current environment for debugging
    logging.debug(f"Initial DEFECTS4J_HOME: {os.environ.get('DEFECTS4J_HOME', 'Not set')}")
    logging.debug(f"Initial PATH: {os.environ.get('PATH', 'Not set')}")
    
    generate_unit_tests(defects4j_home)