import re
import javalang
import javalang.tree
from typing import List, Optional

from LLMTests.utils.logging_utils import LoggerSetup

from LLMTests.extractMethods.method_data import ConstructorData, MethodData

logger_setup = LoggerSetup(logger_name="MethodExtractor")
logger = logger_setup.get_logger()
logger.setLevel("INFO")

class MethodExtractor:
    """Enhanced class for extracting methods and constructors from Java and Kotlin files."""
    
    def __init__(self):
        self.public_methods: List[MethodData] = []
        self.private_methods: List[MethodData] = []
        self.constructors: List[ConstructorData] = []
        self.imports: str = ""
        self.package: str = ""
        self.source_code: str = ""
        self.class_name: str = ""
        
        # Regular expression patterns
        self.patterns = {
            'loop_end': re.compile(r'^\s*}'),
            'javadoc_start': re.compile(r'^\s*/\*\*'),
            'javadoc_end': re.compile(r'^\s*\*/'),
            'start_loop': re.compile(r'{'),
            'import': re.compile(r'^\s*import\s+'),
            'package': re.compile(r'^\s*package\s+([^;]+);'),
            'kotlin_method': re.compile(r"^\s*[a-z]*\s*fun\s+(\w+)\s*\([^)]*\)\s*:"),
            'class_declaration': re.compile(r'^\s*(?:public\s+)?class\s+(\w+)')
        }

    def extract_methods_java(self, file_path: str) -> None:
        """
        Extract methods and constructors from a Java file with enhanced context.
        
        Args:
            file_path (str): Path to the Java file
        """
        logger.info(f"Extracting methods from Java file: {file_path}")
        try:
            with open(file_path, 'r') as file:
                self.source_code = file.read()
        except IOError as e:
            logger.error(f"Error reading file {file_path}: {str(e)}")
            return
        
        # Split code into lines for processing
        lines = self.source_code.split('\n')
        
        # Extract package and imports
        self._extract_package_and_imports(lines)
        
        try:
            # Parse the Java code
            tree = javalang.parse.parse(self.source_code)
        except javalang.parser.JavaSyntaxError as e:
            logger.error(f"Syntax error while parsing Java file {file_path}: {str(e)}")
            # Attempt to remove erroneous block and retry parsing
            self._remove_error_block_and_retry(e)
            return
        except Exception as e:
            logger.error(f"Unexpected error while parsing Java file {file_path}: {str(e)}")
            return
        
        # Process the parsed tree
        self._process_parsed_tree(tree)

    def _handle_partial_parsing(self, lines: List[str]) -> None:
        """Handle partial parsing of the source code in case of syntax errors."""
        logger.info("Attempting partial parsing of the source code")
        for i, line in enumerate(lines):
            if self._is_balanced_block_comment(lines[:i+1]):
                try:
                    partial_tree = javalang.parse.parse('\n'.join(lines[:i+1]))
                    for path, node in partial_tree:
                        if isinstance(node, javalang.tree.MethodDeclaration):
                            self._process_method_declaration(node, lines)
                        elif isinstance(node, javalang.tree.ConstructorDeclaration):
                            self._process_constructor_declaration(node, lines)
                except javalang.parser.JavaSyntaxError:
                    continue
                except Exception as e:
                    logger.error(f"Unexpected error during partial parsing: {str(e)}")
                    break

    def _is_balanced_block_comment(self, lines: List[str]) -> bool:
        """Check if block comments are balanced in the given lines."""
        open_comments = sum(line.count('/*') for line in lines)
        close_comments = sum(line.count('*/') for line in lines)
        return open_comments == close_comments

    def _extract_package_and_imports(self, lines: List[str]) -> None:
        """Extract package declaration and imports from code."""
        logger.debug("Extracting package and imports")
        for line in lines:
            if package_match := self.patterns['package'].match(line):
                self.package = package_match.group(1).strip()
                logger.debug(f"Extracted package: {self.package}")
            elif self.patterns['import'].match(line):
                self.imports += line + '\n'
                logger.debug(f"Extracted import: {line.strip()}")

    def _process_method_declaration(self, node: javalang.tree.MethodDeclaration, lines: List[str]) -> None:
        """Process a method declaration node."""
        method_name = node.name
        params = [(param.type.name, param.name) for param in node.parameters]
        return_type = node.return_type.name if node.return_type else "void"
        
        # Build method signature
        param_str = ', '.join(f"{type_name} {name}" for type_name, name in params)
        signature = f"{return_type} {method_name}({param_str})"
        
        # Find method in source code and extract body and comments
        method_body, method_start_line = self._find_method_body(lines, node.position.line -1)
        comments = self._find_method_comments(lines, method_start_line)
        
        # Create MethodData object with enhanced information
        method_data = MethodData(
            signature=signature,
            comments=comments,
            method_body=method_body,
            source_code=self.source_code,
            class_name=self.class_name,
            package=self.package
        )
        
        # Add to appropriate list based on visibility
        if node.modifiers and "private" in node.modifiers:
            self.private_methods.append(method_data)
            logger.debug(f"Added private method: {signature}")
        else:
            self.public_methods.append(method_data)
            logger.debug(f"Added public method: {signature}")

    def _process_constructor_declaration(self, node: javalang.tree.ConstructorDeclaration, lines: List[str]) -> None:
        """Process a constructor declaration node."""
        params = [(param.type.name, param.name) for param in node.parameters]
        param_str = ', '.join(f"{type_name} {name}" for type_name, name in params)
        signature = f"public {node.name}({param_str})"
        
        constructor_body, constructor_start_line = self._find_method_body(lines, node.position.line -1)
        
        constructor = ConstructorData(
            signature=signature,
            constructor_body=constructor_body,
            class_name=self.class_name,
            package=self.package
        )
        self.constructors.append(constructor)
        logger.debug(f"Added constructor: {signature}")

    def _find_method_body(self, lines: List[str], start_line: int) -> tuple[str, int]:
        """Find and extract method body from code."""
        logger.debug(f"Finding method body starting from line: {start_line + 1}")
        body = []
        brace_count = 0
        found_start = False

        for i in range(start_line, len(lines)):
            line = lines[i]
            if not found_start:
                if '{' in line:
                    found_start = True
                    brace_count += line.count('{')
                    body.append(line)
                else:
                    continue
            else:
                body.append(line)
                brace_count += line.count('{')
                brace_count -= line.count('}')

                if brace_count == 0:
                    logger.debug(f"Method body found: {body}")
                    return '\n'.join(body), start_line
        logger.warning(f"Method body not found for start line: {start_line + 1}")
        return "", start_line

    def _find_method_comments(self, lines: List[str], method_start_line: int) -> str:
        """Find and extract method comments from code."""
        logger.debug(f"Finding comments for method starting at line: {method_start_line + 1}")
        comments = []
        i = method_start_line - 1
        in_javadoc = False

        # Search backwards from the method start line
        while i >= 0:
            line = lines[i].strip()

            if self.patterns['javadoc_end'].match(line):
                in_javadoc = True
                comments.insert(0, line)
            elif self.patterns['javadoc_start'].match(line):
                in_javadoc = False
                comments.insert(0, line)
                return '\n'.join(comments)
            elif in_javadoc:
                comments.insert(0, line)
            elif line.startswith("//"):
                comments.insert(0, line)
            elif not in_javadoc and line == "":
                # Stop searching if we encounter a blank line outside a Javadoc block
                break
            i -= 1

        logger.warning(f"Comments not found for method starting at line: {method_start_line + 1}")
        return ""

    def build_body(self, code: List[str], index: int) -> str:
        """Build method/constructor body with improved formatting."""
        body = []
        brace_count = 0
        
        # Handle opening brace
        while index < len(code) and '{' not in code[index]:
            index += 1
        if index < len(code):
            body.append(code[index])
            brace_count = code[index].count('{')
            index += 1
        
        # Build body until matching closing brace
        while index < len(code) and brace_count > 0:
            line = code[index]
            body.append(line)
            brace_count += line.count('{') - line.count('}')
            index += 1
            
        return '\n'.join(body)

    def build_comment(self, code: List[str], index: int) -> str:
        """Build method/constructor comments with improved formatting."""
        comments = []
        i = index - 1
        
        while i >= 0 and not bool(self.patterns['javadoc_start'].match(code[i])):
            comments.insert(0, code[i])
            i -= 1
            
        if i >= 0:
            comments.insert(0, code[i])  # Add javadoc start
            
        return '\n'.join(comments)

    def _remove_error_block_and_retry(self, parse_error: javalang.parser.JavaSyntaxError) -> None:
        """Remove the code block containing the syntax error and attempt to reparse."""
        error_line = parse_error.at.position[0]  # Get the line number where the error occurred
        logger.info(f"Syntax error at line {error_line}, attempting to remove the erroneous block and reparse")

        lines = self.source_code.split('\n')
        block_start, block_end = self._find_code_block(lines, error_line - 1)  # Adjusting index
        if block_start is not None and block_end is not None:
            logger.info(f"Removing lines {block_start+1}-{block_end+1}")
            modified_lines = lines[:block_start] + lines[block_end+1:]
            modified_code = '\n'.join(modified_lines)
            try:
                tree = javalang.parse.parse(modified_code)
                logger.info("Successfully parsed after removing the erroneous block")
                self.source_code = modified_code
                self._process_parsed_tree(tree)
            except javalang.parser.JavaSyntaxError as e:
                logger.error(f"Parsing failed after removing block: {str(e)}")
            except Exception as e:
                logger.error(f"Unexpected error during reparsing: {str(e)}")
        else:
            logger.error("Could not identify the erroneous code block to remove")

    def _find_code_block(self, lines: List[str], error_line_index: int) -> (Optional[int], Optional[int]):
        """Find the start and end line numbers of the code block containing the error."""
        stack = []
        for index, line in enumerate(lines):
            if '{' in line:
                stack.append(index)
            if '}' in line and stack:
                start_index = stack.pop()
                if start_index <= error_line_index <= index:
                    return start_index, index
        return None, None

    def _process_parsed_tree(self, tree) -> None:
        """Process the successfully parsed AST to extract methods and constructors."""
        # Extract class name
        class_found = False
        for path, node in tree.filter(javalang.tree.ClassDeclaration):
            self.class_name = node.name
            logger.debug(f"Extracted class name: {self.class_name}")
            class_found = True
            break

        if not class_found:
            logger.warning("No class declaration found in the parsed code")
            return

        # Process each class member
        lines = self.source_code.split('\n')
        for path, node in tree:
            if isinstance(node, javalang.tree.MethodDeclaration):
                self._process_method_declaration(node, lines)
            elif isinstance(node, javalang.tree.ConstructorDeclaration):
                self._process_constructor_declaration(node, lines)

    # Getters remain the same
    def get_public_methods(self) -> List[MethodData]:
        return self.public_methods
    
    def get_private_methods(self) -> List[MethodData]:
        return self.private_methods
    
    def get_constructors(self) -> List[ConstructorData]:
        return self.constructors
    
    def get_imports(self) -> str:
        return self.imports
        
    def get_package(self) -> str:
        return self.package
        
    def get_class_name(self) -> str:
        return self.class_name
        
    def get_source_code(self) -> str:
        return self.source_code