from LLMTests.extractMethods.method_data import MethodData, ConstructorData

class PromptGenerator:
    """Enhanced class for generating prompts for creating unit tests in Java and Kotlin."""

    def __init__(self):
        
        self.java_test_template2 = """
You are an experienced software tester specializing in Java Unit Tests. 

Your task is to create a meaningful JUnit4 test class for the following Java class and method:
{take_a_step_back}
* Package Information:
```java
{package_info}
```
{source_code}
```

* Method to test:
```java
{method_signature}
{method_comments}
{method_body}
```

* Helper methods available:
```java
{helper_methods}
```

* Class constructor(s):
```java
{constructors}
```

* Requirements:
1. Use JUnit4 annotations (@Test, @Before, etc.)
2. Add the appropriate imports (e.g.):
    - org.junit.*;
    - Any required classes from the same package: {package_info}
3. Create a minimal set of test cases for:
    - Normal/expected inputs
    - Edge cases
    - Error conditions
5. Don't write repetitive or redundant test cases. Focus on the most important edge cases and error conditions.
6. Add a test setup if necessary
7. Add corresponding assertions
8. Handle all required exception tests
9. Include the same package declaration as the class being tested

{chain_of_thought}

{n_shot_examples}

The test class should be named {class_name}Test and be public.
Important: Provide complete test implementations including specific test data, method calls, and assertions. 
Do not leave test methods empty or with comments only.
\n* Test:
        """
        
        self.one_shot_example = """
Example test method format:
@Test
public void testMethodWithValidInput() {
    // Arrange
    String input = "test data";
    
    // Act
    String result = testObject.method(input);
    
    // Assert
    assertEquals("Method should process valid input", "expected", result);
}
"""

        self.java_test_template = """
You are an experienced software tester specializing in Java Unit Tests. 
Your task is to create a meaningful JUnit4 test class for the following Java class and method:
{take_a_step_back}
* Package Information:
```java
{package_info}
```
{source_code}
```

* Method to test:
```java
{method_comments}
{method_body}
```
{helper_methods}
{constructors}
* Requirements:
1. Use JUnit4 annotations (@Test, @Before, etc.)
2. Add the appropriate imports (e.g.):
    - org.junit.*;
    - Any required classes from the same package: {package_info}
3. Create a minimal set of test cases for:
    - Normal/expected inputs
    - Edge cases
    - Error conditions
5. Don't write repetitive or redundant test cases. Focus on the most important edge cases and error conditions.
6. Add a test setup if necessary
7. Add corresponding assertions
8. Handle all required exception tests
9. Include the same package declaration as the class being tested
{chain_of_thought}
{n_shot_examples}
The test class should be named {class_name}Test and be public.
Important: Provide complete test implementations including specific test data, method calls, and assertions. 
Do not leave test methods empty or with comments only.
\n* Test:
        """
        
        self.one_shot_example = """
Example test method format:
@Test
public void testMethodWithValidInput() {
    // Arrange
    String input = "test data";
    
    // Act
    String result = testObject.method(input);
    
    // Assert
    assertEquals("Method should process valid input", "expected", result);
}
"""
        
        self.kotlin_test_template = """
        You are an experienced software tester specializing in algorithms. 

        Your task is to create a meaningful JUnit4 test class for the following public Kotlin method:

        {n_shot_examples}

        {take_a_step_back}

        Method signature:
        ```kotlin
        {signature}
        ```

        Method body:
        ```kotlin
        {body}
        ```

        Helper methods:
        ```kotlin
        {helper_methods}
        ```

        KDoc:
        ```kotlin
        {comments}
        ```

        Class constructor(s):
        ```kotlin
        {constructors}
        ```

        Requirements:
        1. Use JUnit4 annotations (@Test, @Before)
        2. Add the appropriate imports (org.junit.Test, org.junit.Assert)
        3. Create test cases for:
           - Normal/expected inputs
           - Edge cases (null, empty, boundary values)
           - Error conditions
        4. Use descriptive test method names
        5. Add a test setup if necessary
        6. Add corresponding assertions
        7. Handle all required exception tests

        {chain_of_thought}

        The test class should be named {class_name}Test and be public.
        """

    def _extract_class_name(self, signature):
        """Extract the class name from the method signature."""
        # Remove access modifiers and return type
        cleaned = ' '.join(signature.split())
        parts = cleaned.split(' ')
        # Find the part before the opening parenthesis
        method_name = next(p for p in parts if '(' in p)
        # Convert to CamelCase if needed
        return method_name.split('(')[0].strip().capitalize()

    def generate_n_shot_examples(self, method_data, n):
        """Generates N-shot examples based on the provided method data."""
        # TODO: Implement logic for selecting and formatting N examples
        # based on method_data and n.
        # Use the information from other theses that already
        # contain test cases for various functions.
        examples = []
        
        examples.append("### N-Shot Examples. \n The tests should look like this: \n")
        for i in range(n):
            # Example: Generate a random test case
            example = f"""
            **Example {i+1}:**
                * Input: [random input data]
                * Expected Output: [expected output based on input data]
            """
            examples.append(example)
        return "\n".join(examples)
    
    # removes everything before the package declaration if something exists before it
   
    def create_java_prompt(self, public_method: MethodData, constructors: list[ConstructorData],
                        private_methods: list[MethodData], *, 
                        n_shots: int = 0, 
                        use_cot: bool = False, 
                        use_tas: bool = False, 
                        use_source_code: bool = True) -> str:
        """Generate an improved prompt for creating a Java unit test."""
        # Get package information
        package_info = public_method.get_package()
        if package_info:
            package_info = f"package {package_info};"
        
        # Format helper methods
        helper_methods_str = "* Helper methods available:\n```java\n".join(method.get_private_method() for method in private_methods).join("\n```\n") if private_methods else ""
        

        # Format constructors
        constructor_str = "\n* Class constructor(s):\n```java\n".join(constructor.get_constructor() for constructor in constructors).join("\n```\n") if constructors and not use_source_code else ""

        # Chain-of-Thought Prompting
        chain_of_thought = """
Let's proceed step-by-step:
1. First, understand the class and method behavior
2. Identify the method's dependencies and required setup
3. Plan test cases for different scenarios
4. Implement the test class with proper assertions
        """ if use_cot else ""

        # Generate N-shot examples
        n_shot_examples = self.one_shot_example if n_shots > 0 else ""
    
        # Take a Step Back Prompting
        take_a_step_back = """
Before we start generating unit tests, let's take a step back and consider:
1. What are the general principles of unit testing?
2. What are the specific challenges of testing this type of method?
3. What are the most important edge cases to consider?
        """ if use_tas else ""
        
        cleaned_source_code = public_method.get_source_code()
        
        source_code = f"""
Full Class to test:
```java
{cleaned_source_code}
        """ if use_source_code else ""

        # Format the prompt using the template
        prompt = self.java_test_template.format(
            take_a_step_back=take_a_step_back,
            package_info=package_info,
            source_code=source_code,
            method_signature=public_method.get_signature(),
            method_comments=public_method.get_comments(),
            method_body=public_method.get_method_body(),
            helper_methods=helper_methods_str,
            constructors=constructor_str,
            chain_of_thought=chain_of_thought,
            n_shot_examples=n_shot_examples,
            class_name=public_method.get_class_name()
        )

        return prompt

    def create_kotlin_prompt(self, public_method: MethodData, constructors: list[ConstructorData],
                        private_methods: list[MethodData], *, n_shots: int = 0,
                        use_cot: bool = False, use_tas: bool = False) -> str:
        """Generate an improved prompt for creating a Kotlin unit test."""
        # Get package information
        package_info = public_method.get_package()
        if package_info:
            package_info = f"package {package_info}"
        
        

        
        # Format helper methods
        helper_methods_str = "* Helper methods available:\n```java\n".join(method.get_private_method() for method in private_methods).join("```") if private_methods else ""

        # Format constructors
        constructor_str = "\n".join(constructor.get_constructor() for constructor in constructors) if constructors else "Default constructor"

        # Chain-of-Thought Prompting
        chain_of_thought = """
        Let's proceed step-by-step:
        1. First, understand the class and method behavior
        2. Identify the method's dependencies and required setup
        3. Plan test cases for different scenarios
        4. Implement the test class with proper assertions
        """ if use_cot else ""

        # Format the prompt using the template
        prompt = self.kotlin_test_template.format(
            package_info=package_info,
            source_code=public_method.get_source_code(),
            method_signature=public_method.get_signature(),
            method_comments=public_method.get_comments(),
            method_body=public_method.get_method_body(),
            helper_methods=helper_methods_str,
            constructors=constructor_str,
            chain_of_thought=chain_of_thought,
            class_name=public_method.get_class_name()
        )

        return prompt

    def create_explanation_prompt(self, failed_tests, source_code):
        """Generate a prompt to explain failed tests."""
        failed_tests_details = '\n'.join(
            f"- Test: {test['test']}\n  Details: {test['details']}" for test in failed_tests
        )

        prompt = f"""
You are an expert software engineer.

The following tests have failed:

{failed_tests_details}

The source code under test is:
```
{source_code}
```

Please provide a detailed explanation of why these tests might be failing and suggest possible fixes.
Format your answer in Markdown and use line breaks for clarity.
            """
        return prompt