import javalang
import javalang.tree
import re
from typing import List, Tuple

from LLMTests.utils.logging_utils import LoggerSetup
logger_setup = LoggerSetup(logger_name="Test Combinator")
logger = logger_setup.get_logger()

def combine_unit_tests(unit_tests: List[Tuple[str, str, object]]) -> str:
    """
    Combines a list of individual unit test strings into a single Java test class.

    Args:
        unit_tests: A list of tuples, where each tuple contains:
                    - The language (e.g., "java")
                    - The unit test code as a string
                    - The MethodData object (not used in this function, but might be useful for future extensions)

    Returns:
        A string containing the combined Java test class code, wrapped in a Markdown code block.
    """
    combined_test_class = ""
    package = None
    imports = set()
    classes = {}  # Dictionary to store test classes by class name

    # Add essential JUnit imports
    imports.add("org.junit.Test")
    imports.add("static org.junit.Assert.*")

    for language, test_code, method_data in unit_tests:
        if language != "java":
            continue  # Skip non-Java tests

        try:
            tree = javalang.parse.parse(test_code)

            # Extract package declaration (only once and from the first valid test code)
            if package is None and tree.package:
                package = tree.package.name

            # Extract ALL imports
            if tree.imports:
                for import_decl in tree.imports:
                    import_path = import_decl.path
                    if import_decl.static:
                        import_path = "static " + import_path
                    if import_decl.wildcard:
                        import_path += ".*"
                    imports.add(import_path)

            # Add specific imports based on usage within the code
            if any(
                keyword in test_code
                for keyword in ["Before", "After", "BeforeClass", "AfterClass"]
            ):
                imports.add("org.junit.Before")
                imports.add("org.junit.After")
                imports.add("org.junit.BeforeClass")
                imports.add("org.junit.AfterClass")

            # Check for assertions and add import if needed
            if "Assert." in test_code:
                imports.add("org.junit.Assert")

            # Extract class declarations and test methods
            for type_decl in tree.types:
                if isinstance(type_decl, javalang.tree.ClassDeclaration):
                    class_name = type_decl.name
                    if class_name not in classes:
                        classes[class_name] = []

                    for member in type_decl.body:
                        if isinstance(member, javalang.tree.MethodDeclaration) and any(
                            isinstance(mod, javalang.tree.Annotation) and mod.name == "Test"
                            for mod in member.annotations
                        ):
                            # Find the start line of the method including annotations
                            method_start = member.position.line - 1

                            # Iterate backwards to find the @Test annotation
                            lines = test_code.splitlines()
                            for i in range(method_start, -1, -1):
                                if "@Test" in lines[i]:
                                    method_start = i
                                    break

                            # Find the end line of the method
                            method_end = None
                            for i in range(method_start, len(lines)):
                                if "}" in lines[i] and lines[i].strip().endswith("}"):
                                    method_end = i
                                    break

                            if method_end is not None:
                                method_code = "\n".join(lines[method_start : method_end + 1])
                                classes[class_name].append(method_code)

        except javalang.parser.JavaSyntaxError as e:
            logger.error(f"Error parsing unit test code: {e}")
            continue
        except Exception as e:
            logger.error(f"An unexpected error occurred: {e}")
            continue

    # Add package declaration (if found)
    if package:
        combined_test_class += f"package {package};\n\n"

    # Add imports to the combined class
    for import_statement in sorted(imports):
        combined_test_class += f"import {import_statement};\n"

    # Add class declarations and test methods
    for class_name, methods in classes.items():
        combined_test_class += f"\npublic class {class_name} {{\n\n"
        for method_code in methods:
            combined_test_class += f"    {method_code}\n\n"
        combined_test_class += "}\n"

    # Wrap the combined test code in a Markdown code block
    combined_test_class = f"```java\n{combined_test_class}\n```"

    return combined_test_class