# execute_tests.py

import json
import traceback
from typing import List, Dict
from pathlib import Path
from .test_executor_mvn import TestExecutor as TestExecutorMvn
from .test_executor import TestExecutor as TestExecutorJunit
from LLMTests.utils.progress_tracker import ProgressTracker
from LLMTests.utils.test_counter import count_total_tests
from LLMTests.utils.logging_utils import LoggerSetup
from LLMTests.testExecution.utils.set_up_test_env import init_test_env
from .test_explainer import generate_test_explanations

logger_setup = LoggerSetup(logger_name="ExecuteTests")
logger = logger_setup.get_logger()

class TestExecutor:
    def __init__(self, test_iteration: str, storage_path: str, progress_callback=None):
        self.test_iteration = test_iteration
        self.storage_path = storage_path
        self.progress_callback = progress_callback
        self.logger = LoggerSetup(logger_name="ExecuteTests").get_logger()
        self.progress = ProgressTracker()
        if self.progress_callback:
            self.progress.set_progress_callback(self.progress_callback)
        self.results = []
        self.total_files = 0
        self.total_tests = 0

    def initialize_test_env(self):
        # Set up the test environment
        target_dir = Path('/home/coder/workspace/white-box-testing/testframeworkwhiteboxmodul/LLMTests/testExecution/javametrics')
        source_project_main = self.storage_path
        storage_path_path = Path(self.storage_path)
        setup_success = init_test_env(target_dir, source_project_main, storage_path_path, iteration_path=self.test_iteration)
        if not setup_success:
            self.logger.error("Failed to set up test environment")
            return False
        return True

    def initialize(self):
        self.logger.info(f"Starting test execution for iteration {self.test_iteration}")
        # Create path object
        self.iteration_path = Path(self.storage_path) / 'unit_tests' / f'{self.test_iteration}_iteration'

        if not self.iteration_path.exists():
            raise FileNotFoundError(f"Iteration directory not found: {self.iteration_path}")

        # Count total tests before execution
        self.total_files, self.total_tests = count_total_tests(self.iteration_path)

        if self.total_tests == 0:
            self.logger.warning("No tests found in the specified directory")
            return False

        # Set total tests in progress tracker
        self.progress.total_methods = self.total_tests
        self.progress.current_method = 0

        # Initial progress update
        self.progress.emit_progress(
            'init',
            f'Starting execution of {self.total_tests} tests from {self.total_files} files',
            {
                'percent': 15,
                'total_tests': self.total_tests,
                'total_files': self.total_files,
                'current_test': 0
            }
        )
        return True

    def execute_new_method(self) -> Dict:
        try:
            executor_mvn = TestExecutorMvn(base_path=self.storage_path, project_path=self.storage_path)
            success, message, test_details = executor_mvn.run_mvn_tests()

            self.logger.info(f"Test execution completed: {success}")

            if not success:
                self.logger.error(f"Test execution failed: {message}")
                self.progress.emit_progress(
                    'error',
                    f"Test execution failed: {message}",
                    {
                        'results': self.results
                    }
                )
                return {'mvn_execution': False}

            # Extract summary details
            total_tests = test_details.get('total_tests', 0)
            passed_tests = test_details.get('passed_tests', 0)
            failed_tests = test_details.get('failed_tests', 0)
            failed_methods = test_details.get('failed_methods', [])
            coverage = test_details.get('coverage', {})
            metrics = test_details.get('metrics', {})

            # Calculate success rate
            success_rate = (passed_tests / total_tests * 100) if total_tests > 0 else 0

            summary = {
                'success': success,
                'message': message,
                'total_tests': total_tests,
                'passed_tests': passed_tests,
                'failed_tests': failed_tests,
                'success_rate': round(success_rate, 2),
                'coverage': coverage
            }

            # Build results list
            self.results = []
            for method in failed_methods:
                self.results.append({
                    'test': method.get('test_name', ''),
                    'status': False,
                    'details': method.get('message', ''),
                    'type': method.get('type', ''),
                    'stack_trace': method.get('stack_trace', '')
                })

            # Include passed tests summary
            if passed_tests > 0:
                self.results.append({
                    'test': 'Passed tests',
                    'status': True,
                    'details': f'{passed_tests} tests passed successfully'
                })
                
            # Test progress update
            self.progress.emit_progress(
                'progress',
                f'Completed {total_tests} tests: {passed_tests} passed, {failed_tests} failed ({success_rate:.1f}% success rate). Generating explanations...',
                {
                    'percent': 50,
                    'total_tests': self.total_tests,
                    'total_files': self.total_files,
                    'current_test': 0
                }
            )


            explanations = {}
            # only generate explanation if there are failed tests and metrics['lines_of_code'] < 1000
            if failed_tests > 0 and metrics.get('lines_of_code', 0) < 1000:
                # Generate explanations for failed tests
                explanations = generate_test_explanations(self.results, base_path=self.storage_path, project_path=self.storage_path)
                if isinstance(explanations, str):
                    explanations = {'explanation': explanations}


            # Emit progress update with explanations on top of existing results
            self.progress.emit_progress(
                'complete',
                f"Completed {total_tests} tests: {passed_tests} passed, {failed_tests} failed ({success_rate:.1f}% success rate)",
                {
                    'explanations': explanations.get('explanation', "No explanation fetched"),
                    'results': self.results,
                    'summary': summary,
                    'metrics': metrics,
                    'coverage': coverage
                }
            )

            # Save test results to JSON file
            results_file = Path(self.storage_path) / 'test_results' / f'{self.test_iteration}_results.json'
            results_file.parent.mkdir(parents=True, exist_ok=True)
            with open(results_file, 'w') as f:
                json.dump({
                    'results': self.results,
                    'explanations': explanations['explanation'] if explanations else {explanations},
                    'summary': summary,
                    'metrics': metrics,
                    'coverage': coverage},
                f)

            combined_results = {
                'mvn_execution': True,
                'explanations': explanations['explanation'] if explanations else {explanations},
                'results': self.results,
                'summary': summary,
                'metrics': metrics,
                'coverage': coverage
            }

            return combined_results

        except Exception as e:
            error_msg = f"Error during new test execution: {str(e)}\n{traceback.format_exc()}"
            self.logger.error(error_msg)
            self.progress.emit_progress(
                'error',
                error_msg,
                {
                    'results': self.results
                }
            )
            return {'mvn_execution': False}

    def execute_old_method(self) -> Dict:
        try:
            executor_junit = TestExecutorJunit(base_path="/home/coder/workspace")
            success_count = 0
            current_method = 0

            # Execute each test file
            test_files = sorted(self.iteration_path.glob('*.json'))

            for test_file in test_files:
                try:
                    self.logger.info(f"Executing testfile: {test_file.name}")
                    
                    # Progress update for starting new file
                    self.progress.emit_progress(
                        'processing',
                        f"Processing test file: {test_file.name}",
                        {
                            'percent': success_count / self.total_tests * 100,
                            'type': 'info',
                            'total_tests': self.total_tests,
                            'current_test': current_method,
                            'current_file': test_file.name
                        }
                    )

                    success, message, test_details = executor_junit.run_single_test_file(str(test_file))
                    
                    # Get metrics from test_details
                    metrics = test_details.get('metrics', {})
                    tests_in_current_file = test_details.get('total_tests', 0)
                    failed_methods = test_details.get('failed_methods', [])

                    # Update progress for each test in the file
                    total_tests_in_file = tests_in_current_file
                    passed_tests_in_file = test_details.get('passed_tests', 0)
                    failed_tests_in_file = test_details.get('failed_tests', 0)
                    current_method += total_tests_in_file

                    # Increment success_count
                    success_count += passed_tests_in_file

                    # Progress update for file completion
                    self.progress.emit_progress(
                        'progress',
                        f"Completed file {test_file.name}: {passed_tests_in_file}/{total_tests_in_file} tests passed",
                        {
                            'type': 'success' if success else 'error',
                            'total_tests': self.total_tests,
                            'current_test': current_method,
                            'file_complete': True,
                            'file_results': {
                                'file': test_file.name,
                                'passed': passed_tests_in_file,
                                'failed': failed_tests_in_file,
                                'total': total_tests_in_file
                            },
                            'results': self.results
                        }
                    )

                    self.results.append({
                        'test': test_file.name,
                        'status': success,
                        'details': message,
                        'test_details': test_details,
                        'metrics': metrics
                    })

                except Exception as e:
                    error_msg = f"Error executing test {test_file.name}: {str(e)}"
                    self.logger.error(error_msg)
                    self.results.append({
                        'test': test_file.name,
                        'status': False,
                        'details': error_msg
                    })

            # Calculate final statistics
            success_rate = (success_count / self.total_tests * 100) if self.total_tests > 0 else 0
            summary = {
                'total_tests': self.total_tests,
                'passed_tests': success_count,
                'failed_tests': self.total_tests - success_count,
                'success_rate': round(success_rate, 2)
            }

            # Final progress update with results
            self.progress.emit_progress(
                'complete',
                f"Completed {self.total_tests} tests: {success_count} passed, {self.total_tests - success_count} failed ({success_rate:.1f}% success rate)",
                {
                    'results': self.results,
                    'summary': summary
                }
            )

            # Generate explanations for failed tests
            explanations = generate_test_explanations(self.results, self.storage_path)

            # Emit progress update with explanations on top of existing results
            self.progress.emit_progress(
                'complete',
                f"Completed {self.total_tests} tests: {success_count} passed, {self.total_tests - success_count} failed ({success_rate:.1f}% success rate)",
                {
                    'explanations': explanations,
                    'results': self.results,
                    'summary': summary
                }
            )

            # Save test results to JSON file
            results_file = Path(self.storage_path) / 'test_results' / f'{self.test_iteration}_results.json'
            results_file.parent.mkdir(parents=True, exist_ok=True)
            with open(results_file, 'w') as f:
                json.dump({'results': self.results, 'summary': summary}, f)

            return {
                'mvn_execution': False,
                'results': self.results,
                'summary': summary,
                'metrics': {},
                'coverage': {}
            }

        except Exception as e:
            error_msg = f"Error during old test execution: {str(e)}\n{traceback.format_exc()}"
            self.logger.error(error_msg)
            self.progress.emit_progress(
                'error',
                error_msg,
                {
                    'results': self.results
                }
            )
            return {
                'mvn_execution': False,
                'results': self.results,
                'summary': {},
                'metrics': {},
                'coverage': {}
            }

    def execute_tests(self) -> Dict:
        if not self.initialize_test_env():
            return {'mvn_execution': False, 'results': []}

        if not self.initialize():
            return {'mvn_execution': False, 'results': []}

        results = self.execute_new_method()
        if not results.get('mvn_execution'):
            self.logger.error("New method failed. Falling back to old method.")
            return self.execute_old_method()
        else:
            return results

def execute_tests(test_iteration: str, storage_path: str, progress_callback=None) -> Dict:
    executor = TestExecutor(test_iteration, storage_path, progress_callback)
    results = executor.execute_tests()
    
    # Fix explanations handling - explanations is already a string
    if 'explanations' in results:
        if isinstance(results['explanations'], dict):
            results['explanations'] = results['explanations'].get('explanation', "No explanation fetched")
        elif not isinstance(results['explanations'], str):
            results['explanations'] = "No explanation fetched"
    else:
        results['explanations'] = "No explanation fetched"
        
    return results
