import subprocess
import os
import re

from LLMTests.utils.logging_utils import LoggerSetup
import xml.etree.ElementTree as ET
from pathlib import Path


logger_setup = LoggerSetup(logger_name="JavaTestRunner")
logger = logger_setup.get_logger()
logger.setLevel('DEBUG')

class CompilationError(Exception):
    """Custom exception for compilation errors."""
    pass

class JavaTestRunner:
    """Class to execute and analyze Java tests using Maven and JaCoCo."""

    def __init__(self, base_dir: str, project_path: str = None):

        self.base_dir = Path(base_dir)
        self.project_test_setup_path = Path('/home/coder/workspace/white-box-testing/testframeworkwhiteboxmodul/LLMTests/testExecution/javametrics')
        self.src_main_java = self.project_test_setup_path / 'src' / 'main' / 'java'
        self.src_test_java = self.project_test_setup_path / 'src' / 'test' / 'java'

        self.maven_output = ''
        self.maven_error = ''
        self.returncode = None
        self.test_results = {
            'total_tests': 0,
            'failures': 0,
            'errors': 0,
            'skipped': 0,
            'failed_methods': [],
            'total_time': 0.0,
            'error_messages': [],
            'test_suite_name': ''
        }

    def run_maven_commands(self):
        """Execute Maven commands to compile, test, and generate coverage reports."""
        maven_command = [
            'mvn',
            'clean',
            'verify',
            '-B',
            '-V'
        ]
        
        process = subprocess.Popen(
            maven_command,
            cwd=str(self.project_test_setup_path),
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
            shell=False,
            env=dict(os.environ, JAVA_HOME=os.environ.get('JAVA_HOME', ''))
        )

        stdout, stderr = process.communicate(timeout=120)
        self.maven_output = stdout.decode('utf-8')
        self.maven_error = stderr.decode('utf-8')
        self.returncode = process.returncode
        
        #logger.debug(f"!!!!Maven output: {self.maven_output}")

        # Remove exception raising and log a warning instead
        if self.returncode != 0:
            logger.warning(f"Maven execution completed with return code {self.returncode}. Some tests may have failed.")
            # Proceed to parse test results and coverage
        
        return self.maven_output, self.maven_error, self.returncode

    def parse_test_results(self):
        """Parse all test results from Surefire reports"""
        report_dir = self.project_test_setup_path / 'target' / 'surefire-reports'
        test_results = {
            'total_tests': 0,
            'failures': 0,
            'errors': 0,
            'skipped': 0,
            'failed_methods': [],
            'total_time': 0.0,
            'error_messages': [],
            'test_suite_name': ''
        }
        for report_file in report_dir.glob('*.xml'):
            try:
                tree = ET.parse(report_file)
                root = tree.getroot()
                
                # Safely get and convert values with validation
                tests = root.get('tests')
                failures = root.get('failures')
                errors = root.get('errors')
                skipped = root.get('skipped')
                time = root.get('time')
                
                # Validate and convert numeric values
                try:
                    test_results['total_tests'] += int(tests) if tests is not None else 0
                    test_results['failures'] += int(failures) if failures is not None else 0
                    test_results['errors'] += int(errors) if errors is not None else 0
                    test_results['skipped'] += int(skipped) if skipped is not None else 0
                    test_results['total_time'] += float(time) if time is not None else 0.0
                except ValueError as e:
                    logger.error(f"Invalid numeric value in report {report_file}: {e}")
                    continue

                test_results['test_suite_name'] = root.get('name', 'Unknown Test Suite')
                
                # Process test cases
                for testcase in root.iter('testcase'):
                    if testcase is None:
                        continue
                        
                    failures = testcase.findall('failure')  # Handle multiple failures
                    for failure in failures:
                        if failure is not None:
                            failure_data = {
                                'test_name': testcase.get('name', 'Unknown Test'),
                                'message': failure.get('message', 'No message provided'),
                                'type': failure.get('type', 'Unknown Type'),
                                'stack_trace': failure.text if failure.text else 'No stack trace available'
                            }
                            test_results['failed_methods'].append(failure_data)
                            if failure.text:
                                test_results['error_messages'].append(failure.text)
                                
            except ET.ParseError as e:
                logger.error(f"Failed to parse XML file {report_file}: {e}")
                continue
            except Exception as e:
                logger.error(f"Unexpected error processing {report_file}: {e}")
                continue

        # Ensure all values are non-negative
        for key in ['total_tests', 'failures', 'errors', 'skipped']:
            test_results[key] = max(0, test_results[key])
        test_results['total_time'] = max(0.0, test_results['total_time'])
        
        #logger.info(f"Parsed Test results: {test_results}")
        
        return test_results

    def parse_code_coverage(self):
        """Parse the JaCoCo XML report to extract coverage metrics."""
        jacoco_report_path = self.project_test_setup_path / 'target' / 'site' / 'jacoco' / 'jacoco.xml'
        if not jacoco_report_path.exists():
            logger.error(f"JaCoCo report not found at {jacoco_report_path}")
            
            code_coverage = {
                'line_coverage': 0.0,
                'branch_coverage': 0.0
            }
            return code_coverage

        tree = ET.parse(jacoco_report_path)
        root = tree.getroot()

        total_lines = 0
        covered_lines = 0
        total_branches = 0
        covered_branches = 0

        # Iterate over all class elements
        for clazz in root.findall('.//class'):
            line_counter = clazz.find("./counter[@type='LINE']")
            if line_counter is not None:
                covered = int(line_counter.get('covered', '0'))
                missed = int(line_counter.get('missed', '0'))
                total_lines += covered + missed
                covered_lines += covered

            branch_counter = clazz.find("./counter[@type='BRANCH']")
            if branch_counter is not None:
                covered = int(branch_counter.get('covered', '0'))
                missed = int(branch_counter.get('missed', '0'))
                total_branches += covered + missed
                covered_branches += covered

        code_coverage = {
            'line_coverage': self._calculate_coverage(covered_lines, total_lines - covered_lines),
            'branch_coverage': self._calculate_coverage(covered_branches, total_branches - covered_branches)
        }
        return code_coverage

    def _calculate_coverage(self, covered, missed):
        """Helper method to calculate coverage percentage."""
        total = covered + missed
        if total == 0:
            return 0.0 * 100
        else:
            return round(covered / total * 100, 2)

    def calculate_other_metrics(self):
        """
        Calculate metrics like lines of code, number of assertions,
        number of test cases, test item LOC, number of failed test cases,
        and pass@1 metric, similar to metric.py.
        """
        metrics = {}
        # Calculate lines of code in test files
        metrics['lines_of_code'] = self.count_total_lines(self.src_test_java)
        # Calculate lines of code in source files
        metrics['test_item_loc'] = self.count_total_lines(self.src_main_java)
        # Count number of assertions and test cases
        metrics['number_of_assertions'], metrics['number_of_testcases'] = self.count_assertions_and_tests(self.src_test_java)
        # Parse test results for failed tests
        failed_methods = self.parse_surefire_reports()
        metrics['number_failed_testcases'] = len(failed_methods)
        # Calculate pass@1 metric
        metrics['pass@1'] = 1 if metrics['number_failed_testcases'] == 0 else 0
        # Log summary
        
        
        logger.info(f"Total source LOC: {metrics['test_item_loc']}")
        logger.info(f"Total test LOC: {metrics['lines_of_code']}")
        logger.info(f"Total test cases: {metrics['number_of_testcases']}")
        return metrics

    def count_total_lines(self, directory):
        total_lines = 0
        for java_file in directory.rglob('*.java'):
            try:
                with java_file.open('r', encoding='utf-8') as f:
                    total_lines += sum(1 for _ in f)
            except Exception as e:
                logger.error(f"Error reading file {java_file}: {str(e)}")
        return total_lines

    def count_assertions_and_tests(self, directory):
        total_assertions = 0
        total_tests = 0
        for java_file in directory.rglob('*.java'):
            try:
                with java_file.open('r', encoding='utf-8') as f:
                    content = f.read()
                    total_assertions += len(re.findall(r'\bassert\w*\b', content))
                    total_tests += content.count('@Test')
            except Exception as e:
                logger.error(f"Error processing file {java_file}: {str(e)}")
        return total_assertions, total_tests

    def parse_surefire_reports(self):
        """Parse Surefire reports to collect failed test methods."""
        report_dir = self.project_test_setup_path / 'target' / 'surefire-reports'
        failed_methods = []
        for report_file in report_dir.glob('TEST-*.xml'):
            try:
                tree = ET.parse(report_file)
                root = tree.getroot()
                for testcase in root.iter('testcase'):
                    failure = testcase.find('failure')
                    if failure is not None:
                        method_name = testcase.get('name')
                        failed_methods.append(method_name)
            except Exception as e:
                logger.error(f"Error parsing report {report_file}: {str(e)}")
        return failed_methods