# test_executor.py

import subprocess
import os
import re
import json
import tempfile
import logging
import traceback
from pathlib import Path
from typing import List, Optional, Tuple, Dict
import shutil

from .test_failure import TestFailure
from .test_result import TestResult
from .exceptions import CompilationError

logger = logging.getLogger(__name__)

class TestExecutor:
    """Class for executing individual Java/Kotlin unit tests."""
    
    def __init__(self, base_path: str = "/home/coder/workspace", project_path: str = None):
        """Initialize the test executor with necessary paths."""
        self.base_path = Path(base_path)
        self.junit_path = self.base_path / "white-box-testing/testframeworkwhiteboxmodul/LLMTests/runtimes/junit-4.13.2.jar"
        self.hamcrest_path = self.base_path / "white-box-testing/testframeworkwhiteboxmodul/LLMTests/runtimes/hamcrest-core-1.3.jar"
        
        self.failure_pattern = re.compile(
            r'(?P<test_name>\w+)(?:\([\w.]+\))?\s*[:]\s*'
            r'(?P<message>.*?)\s*\n\s*at\s+'
            r'(?P<stack_trace>.*?)(?=\n\n|\Z)',
            re.DOTALL
        )
        self._verify_dependencies()
    
    def _verify_dependencies(self):
        """Verify that all required dependencies are available."""
        if not self.junit_path.exists():
            raise FileNotFoundError(f"JUnit JAR not found at {self.junit_path}")
        if not self.hamcrest_path.exists():
            raise FileNotFoundError(f"Hamcrest JAR not found at {self.hamcrest_path}")

    def _extract_package_name(self, code: str) -> Optional[str]:
        """Extract package name from Java code."""
        match = re.search(r'package\s+([\w.]+);', code)
        return match.group(1) if match else None

    def _extract_class_name(self, code: str) -> Optional[str]:
        """Extract class name from Java code."""
        patterns = [
            r'public\s+class\s+(\w+)',
            r'class\s+(\w+)',
            r'public\s+class\s+(\w+Test)',
            r'class\s+(\w+Test)'
        ]
        
        for pattern in patterns:
            if match := re.search(pattern, code):
                return match.group(1)
        return None

    def run_single_test_file(self, test_path: str) -> Tuple[bool, str, Dict]:
        """Execute a single test file from a JSON file and return detailed results."""
        try:
            logger.debug(f"Running single test file: {test_path}")
            with open(test_path, 'r') as f:
                test_data = json.load(f)

            test_code = test_data['code']['testCode']
            source_code = test_data['code'].get('sourceCode')
            
            if not source_code:
                logger.error("No source code provided in test data")
                return False, "No source code provided in test data", {}

            package_name = self._extract_package_name(test_code)
            source_class_name = self._extract_class_name(source_code)
            test_class_name = self._extract_class_name(test_code)

            if not source_class_name or not test_class_name:
                logger.error("Could not extract class names")
                return False, "Could not extract class names", {}
            
            # Extract individual test methods and their line numbers
            test_methods = self._extract_test_methods(test_code)
            # logger.debug(f"Found test methods: {test_methods}")

            # Count tests in this file
            test_count = len(test_methods)
            if test_count == 0:
                logger.warning("No @Test annotations found in test file")
                return False, "No tests found", {"tests_in_file": 0}

            with tempfile.TemporaryDirectory() as temp_dir:
                package_path, source_file, test_file = self._prepare_test_environment(
                    test_code, source_code, temp_dir, package_name, source_class_name, test_class_name
                )
                
                classpath = f"{temp_dir}{os.pathsep}{self.junit_path}{os.pathsep}{self.hamcrest_path}"
                
                try:
                    self.compile_java([source_file, test_file], classpath)
                except CompilationError as ce:
                    logger.error(f"Compilation Error: {str(ce)}")
                    return False, f"Compilation Error:\n{str(ce)}", {
                        'error': str(ce),
                        'error_type': 'compilation',
                        'test_code': test_code
                    }

                test_results = self._run_test(test_class_name, classpath, package_name)
                
                # Enhanced results with detailed test information
                detailed_results = self._create_detailed_results(
                    test_results, 
                    test_methods, 
                    test_code
                )
                
                results_dict = {
                    'success': test_results.success,
                    'total_tests': test_results.total_tests,
                    'passed_tests': test_results.passed_tests,
                    'detailed_results': detailed_results,  # This contains individual test results
                    'test_methods': test_methods,
                    'failures': [vars(f) for f in test_results.failures],
                    'errors': test_results.errors,
                    'stdout': test_results.stdout,
                    'stderr': test_results.stderr,
                    'tests_in_file': test_count,
                    'test_code': test_code
                }

                # Ensure Maven is available
                if not shutil.which("mvn"):
                    raise EnvironmentError("Maven (mvn) command not found. Please ensure Maven is installed and available in the PATH.")
                
                # Calculate metrics using calculate_metrics from metric.py
                #metrics = calculate_metrics(source_code, test_code, lang="java")
                # Include metrics in the results dictionary
                #results_dict['metrics'] = metrics

                return test_results.success, test_results.stdout, results_dict

        except Exception as e:
            error_msg = f"Test Execution Error:\n{str(e)}\n{traceback.format_exc()}"
            logger.error(error_msg)
            return False, error_msg, {'error': str(e)}

    def run_single_test_code(self, test_code: str, source_code: str) -> Tuple[bool, str, Dict]:
        """Execute a single test from code strings and return detailed results."""
        try:
            logger.debug("Running single test from code strings")
            
            if not source_code:
                logger.error("No source code provided")
                return False, "No source code provided", {}

            package_name = self._extract_package_name(test_code)
            source_class_name = self._extract_class_name(source_code)
            test_class_name = self._extract_class_name(test_code)

            if not source_class_name or not test_class_name:
                logger.error("Could not extract class names")
                return False, "Could not extract class names", {}

            # Extract individual test methods and their line numbers
            test_methods = self._extract_test_methods(test_code)
            logger.debug(f"Found test methods: {test_methods}")

            # Count tests in this code
            test_count = len(test_methods)
            if test_count == 0:
                logger.warning("No @Test annotations found in test code")
                return False, "No tests found", {"tests_in_code": 0}

            with tempfile.TemporaryDirectory() as temp_dir:
                package_path, source_file, test_file = self._prepare_test_environment(
                    test_code, source_code, temp_dir, package_name, source_class_name, test_class_name
                )
                
                classpath = f"{temp_dir}{os.pathsep}{self.junit_path}{os.pathsep}{self.hamcrest_path}"
                
                try:
                    self.compile_java([source_file, test_file], classpath)
                except CompilationError as ce:
                    logger.error(f"Compilation Error: {str(ce)}")
                    
                    results_dict = {
                        'success': False,
                        'total_tests': 0,
                        'passed_tests': 0,
                        'detailed_results': [],
                        'test_methods': test_methods,
                        'failures': [],
                        'errors': [str(ce)],
                        'stdout': '',
                        'stderr': '',
                        'tests_in_code': test_count,
                        'test_code': test_code
                    }
                    
                    return False, results_dict, str(ce)

                test_results = self._run_test(test_class_name, classpath, package_name)
                
                # Enhanced results with detailed test information
                detailed_results = self._create_detailed_results(
                    test_results, 
                    test_methods, 
                    test_code
                )
                
                results_dict = {
                    'success': test_results.success,
                    'total_tests': test_results.total_tests,
                    'passed_tests': test_results.passed_tests,
                    'detailed_results': detailed_results,
                    'test_methods': test_methods,
                    'failures': [vars(f) for f in test_results.failures],
                    'errors': test_results.errors,
                    'stdout': test_results.stdout,
                    'stderr': test_results.stderr,
                    'tests_in_code': test_count,
                    'test_code': test_code
                }

                return test_results.success, results_dict, ""

        except Exception as e:            
            error_msg = f"Test Execution Error:\n{str(e)}\n{traceback.format_exc()}"
            logger.error(f"Test Execution Error from Junitexecution: {error_msg}")
            return False, None, error_msg

    def _extract_test_methods(self, test_code: str) -> List[Dict]:
        """Extract test methods and their line numbers from the test code."""
        test_methods = []
        lines = test_code.split('\n')
        current_method = None
        start_line = None
        
        for i, line in enumerate(lines, 1):
            # Look for @Test annotation
            if '@Test' in line:
                # Look for the method declaration in the next lines
                for j, next_line in enumerate(lines[i:], i):
                    if 'public' in next_line and 'void' in next_line and '(' in next_line:
                        method_name = re.search(r'void\s+(\w+)\s*\(', next_line)
                        if method_name:
                            current_method = method_name.group(1)
                            start_line = i
                            break
            
            # Look for end of method if we're tracking one
            elif current_method and '}' in line:
                test_methods.append({
                    'name': current_method,
                    'start_line': start_line,
                    'end_line': i,
                    'code': '\n'.join(lines[start_line-1:i])
                })
                current_method = None
                
        return test_methods

    def _create_detailed_results(self, test_results, test_methods, test_code) -> List[Dict]:
        """Create detailed results for each test method."""
        detailed_results = []
        
        # Create a map of test method names to their details
        # method_map = {method['name']: method for method in test_methods}
        
        # Track which tests passed (all tests not in failures)
        failed_tests = {failure.test_name for failure in test_results.failures}
        
        for method in test_methods:
            test_name = method['name']
            is_failed = test_name in failed_tests
            
            result = {
                'test_name': test_name,
                'status': 'failed' if is_failed else 'passed',
                'line_range': (method['start_line'], method['end_line']),
                'test_code': method['code'],
            }
            
            # Add failure details if the test failed
            if is_failed:
                failure = next((f for f in test_results.failures if f.test_name == test_name), None)
                if failure:
                    result.update({
                        'failure_message': failure.message,
                        'failure_type': failure.failure_type,
                        'stack_trace': failure.stack_trace,
                    })
                    
                    # Try to extract assertion failure line number from stack trace
                    line_match = re.search(r':(\d+)\)', failure.stack_trace)
                    if line_match:
                        result['failure_line'] = int(line_match.group(1))
            
            detailed_results.append(result)
            
        return detailed_results

    def _prepare_test_environment(self, test_code: str, source_code: str, temp_dir: str,
                                  package_name: Optional[str], source_class_name: str,
                                  test_class_name: str) -> Tuple[str, str, str]:
        """Prepare test environment with source and test files."""
        if package_name:
            package_path = os.path.join(temp_dir, *package_name.split('.'))
            os.makedirs(package_path, exist_ok=True)
        else:
            package_path = temp_dir

        # Create source file with the exact class name
        source_file = os.path.join(package_path, f"{source_class_name}.java")
        test_file = os.path.join(package_path, f"{test_class_name}.java")

        # Write files with correct encoding
        with open(source_file, 'w', encoding='utf-8') as f:
            f.write(source_code)
        with open(test_file, 'w', encoding='utf-8') as f:
            f.write(test_code)

        return package_path, source_file, test_file

    def compile_java(self, files: List[str], classpath: str) -> None:
        """Compile Java files with enhanced error handling."""
        compile_cmd = ["javac", "-cp", classpath, *files]
        
        process = subprocess.run(compile_cmd, capture_output=True, text=True)
        
        if process.returncode != 0:
            raise CompilationError(process.stderr)

    def _run_test(self, test_class_name: str, classpath: str, package_name: Optional[str]) -> TestResult:
        """Run Java test and process results."""
        full_class_name = f"{package_name}.{test_class_name}" if package_name else test_class_name

        run_cmd = [
            "java",
            "-cp", classpath,
            "org.junit.runner.JUnitCore",
            full_class_name
        ]
        
        process = subprocess.run(run_cmd, capture_output=True, text=True)
        
        # Initialize variables
        total_tests = 0
        passed_tests = 0
        failures = []
        has_failures = False

        output_lines = process.stdout.splitlines()
        failure_block = []
        in_failure_block = False

        for line in output_lines:
            # logger.debug(f"Processing line: {line}")

            # Check for test summary line
            if line.startswith("Tests run:"):
                summary_match = re.search(r'Tests run: (\d+),\s+Failures: (\d+)', line)
                if summary_match:
                    total_tests = int(summary_match.group(1))
                    failure_count = int(summary_match.group(2))
                    passed_tests = total_tests - failure_count
                    has_failures = failure_count > 0
                    logger.debug(f"Total tests: {total_tests}, Failures: {failure_count}, Passed: {passed_tests}")
                    
            # if line.startswith("OK") extract the total number of tests from the line 'OK (8 tests)'
            if line.startswith("OK"):
                total_tests = int(re.search(r'\d+', line).group())
                passed_tests = total_tests
                failure_count = 0
                has_failures = False
                logger.debug(f"Total tests: {total_tests}, Failures: {failure_count}, Passed: {passed_tests}")

            # Detect start of failure block
            if line.startswith("There was ") or line.startswith("There were "):
                in_failure_block = True
                continue

            # Collect failure details
            if in_failure_block:
                if line.strip() == '':
                    # Empty line indicates end of failure block
                    in_failure_block = False
                    continue
                failure_block.append(line)

        # Parse collected failure details
        failure_text = '\n'.join(failure_block)
        failure_matches = re.finditer(
            r'\d+\) (?P<test_name>\w+)\(.*?\)\n(?P<failure_message>.*?)(?=\n\d+\)|\Z)',
            failure_text,
            re.DOTALL
        )

        for match in failure_matches:
            test_name = match.group('test_name')
            message_and_trace = match.group('failure_message').strip()
            parts = message_and_trace.split('\n', 1)
            message = parts[0]
            stack_trace = parts[1] if len(parts) > 1 else ''
            
            failure = TestFailure(
                test_name=test_name,
                failure_type='JUnit Test Failure',
                message=message,
                stack_trace=stack_trace
            )
            failures.append(failure)
            logger.info(f"Test failure recorded: {test_name} - {message}")

        # Determine overall success status
        success = not has_failures and not process.stderr

        if not success:
            logger.warning(f"Test execution failed for {test_class_name}")
            # logger.debug(f"Stdout: {process.stdout}")
            # logger.debug(f"Stderr: {process.stderr}")
            # logger.debug(f"Failures: {failures}")

        return TestResult(
            test_name=test_class_name,
            success=success,
            total_tests=total_tests,
            passed_tests=passed_tests,
            execution_time=0.0,
            failures=failures,
            errors=[process.stderr] if process.stderr else [],
            stdout=process.stdout,
            stderr=process.stderr,
            source_code_context=None
        )