# test_executor.py

import logging
import traceback
from pathlib import Path
from typing import Tuple, Dict
from .java_test_runner import JavaTestRunner

logger = logging.getLogger(__name__)

class TestExecutor:
    """Class for executing individual Java unit tests."""

    def __init__(self, base_path: str = "/home/coder/workspace", project_path: str = None):
        """Initialize the test executor with the base path."""
        # Set up the test environment

        self.base_path = Path(base_path)
        self.project_path = Path(project_path) if project_path else self.base_path

    def run_mvn_tests(self) -> Tuple[bool, str, Dict]:
        """Execute a maven test run and return detailed results."""
        try:
            runner = JavaTestRunner(str(self.base_path), project_path=str(self.project_path))
   
            maven_output, maven_error, returncode = runner.run_maven_commands()

            try: 
                test_results = runner.parse_test_results()
                logger.info("The test results are really: " + str(test_results))
            except Exception as e:
                test_results = {
                    'total_tests': 0,
                    'failures': 0,
                    'errors': 0,
                    'skipped': 0,
                    'line_coverage': 0.0,
                    'failed_methods': [],
                    'total_time': 0.0,
                    'error_messages': [],
                    'test_suite_name': ''
                }
                logger.error(f"Error parsing test results: {str(e)}")
                logger.error(traceback.format_exc())
            
            try:
                logger.info("Parsing code coverage")
                code_coverage = runner.parse_code_coverage()
                other_metrics = runner.calculate_other_metrics()
            except Exception as e:
                code_coverage = {
                    'line_coverage': 0.0,
                    'branch_coverage': 0.0,
                    'complexity': 0.0,
                    'methods': 0,
                    'statements': 0,
                    'classes': 0
                }
                other_metrics = {}
                logger.error(f"Error parsing code coverage: {str(e)}")
                logger.error(traceback.format_exc())

            success = runner.maven_error == '' and runner.returncode == 0
            results_dict = {
                'success': success,
                'total_tests': test_results['total_tests'],
                'passed_tests': test_results['total_tests'] - test_results['failures'] - test_results['errors'],
                'failed_tests': test_results['failures'] + test_results['errors'],
                'failed_methods': test_results['failed_methods'],
                'coverage': code_coverage,
                'metrics': other_metrics,
                'maven_output': runner.maven_output,
                'maven_error': runner.maven_error,
                #'test_code': test_code
            }
            logger.debug(f"#####Test results: {results_dict}")
            return success, runner.maven_output, results_dict

        except Exception as e:
            error_msg = f"Test Execution Error:\n{str(e)}\n{traceback.format_exc()}"
            logger.error(f"Test Execution Error while sending stuff: {error_msg}")

            return False, error_msg, {'error': str(e)}
