import shutil
import json
from typing import Tuple, List, Dict, Optional
from pathlib import Path
import xml.etree.ElementTree as ET  # For parsing XML files
from LLMTests.utils.logging_utils import LoggerSetup

logger_setup = LoggerSetup(logger_name="SetUpTestEnv")
logger = logger_setup.get_logger()
logger.setLevel('INFO')

def clean_target_directory(target_dir: Path) -> Tuple[bool, str]:
    """Cleans the target test directory."""
    try:
        if target_dir.exists() and target_dir.is_dir():
            shutil.rmtree(target_dir)
            logger.info(f"Cleaned target directory: {target_dir}")
        else:
            logger.info(f"Target directory does not exist, no need to clean: {target_dir}")
        return True, ""
    except Exception as e:
        logger.error(f"Failed to clean target directory: {e}")
        return False, str(e)

def _find_pom_file(project_path: Optional[str]) -> Path:
    """Search for a pom.xml file in the project directory."""
    base_pom = Path('/home/coder/workspace/white-box-testing/testframeworkwhiteboxmodul/LLMTests/runtimes/pom.xml')

    if project_path:
        project_dir = Path(project_path)
        if project_dir.is_file() and project_dir.name == 'pom.xml':
            return project_dir
        elif project_dir.is_dir():
            pom_files = list(project_dir.rglob('pom.xml'))
            if pom_files:
                return pom_files[0]
    # Return base pom.xml if none found
    return base_pom

import xml.etree.ElementTree as ET  # For parsing XML files

def adjust_pom_file(pom_file_path: Path, new_group_id: str, new_artifact_id: Optional[str] = None) -> Tuple[bool, str]:
    """
    Adjusts the pom.xml file to have the correct groupId and artifactId.

    :param pom_file_path: Path to the pom.xml file to adjust.
    :param new_group_id: New groupId to set in the pom.xml.
    :param new_artifact_id: (Optional) New artifactId to set in the pom.xml.
    :return: Tuple indicating success and error message if any.
    """
    try:
        # Register namespace to avoid ns0 prefixes
        ET.register_namespace('', 'http://maven.apache.org/POM/4.0.0')

        # Parse the XML file
        tree = ET.parse(pom_file_path)
        root = tree.getroot()

        # Define the namespace mapping
        ns = {'m': 'http://maven.apache.org/POM/4.0.0'}

        # Update the groupId
        group_id_elem = root.find('m:groupId', ns)
        if group_id_elem is not None:
            group_id_elem.text = new_group_id
        else:
            # If groupId is not found, it might be inherited; create it
            group_id_elem = ET.SubElement(root, 'groupId')
            group_id_elem.text = new_group_id

        # Update the artifactId if provided
        if new_artifact_id:
            artifact_id_elem = root.find('m:artifactId', ns)
            if artifact_id_elem is not None:
                artifact_id_elem.text = new_artifact_id
            else:
                # Create artifactId element if it doesn't exist
                artifact_id_elem = ET.SubElement(root, 'artifactId')
                artifact_id_elem.text = new_artifact_id

        # Write the changes back to pom.xml
        tree.write(pom_file_path, encoding='UTF-8', xml_declaration=True)
        logger.info(f"Adjusted pom.xml file at {pom_file_path}")
        return True, ""
    except Exception as e:
        logger.error(f"Failed to adjust pom.xml: {e}")
        return False, str(e)


def extract_package_name(java_files: List[Path]) -> Optional[str]:
    """
    Extracts the package name from Java files.

    :param java_files: List of Java source files.
    :return: The package name if found, otherwise None.
    """
    for java_file in java_files:
        with open(java_file, 'r') as f:
            for line in f:
                line = line.strip()
                if line.startswith('package '):
                    package_name = line[len('package '):].rstrip(';')
                    logger.debug(f"Extracted package name '{package_name}' from {java_file}")
                    return package_name
    logger.warning("No package declaration found in Java files.")
    return None

def create_maven_structure(project_test_setup_path: Path, project_path: Optional[str] = None) -> Tuple[bool, str]:
    """Creates Maven project structure and adjusts the pom.xml."""
    try:
        src_main = project_test_setup_path / 'src' / 'main' / 'java'
        src_test = project_test_setup_path / 'src' / 'test' / 'java'
        src_main.mkdir(parents=True, exist_ok=True)
        src_test.mkdir(parents=True, exist_ok=True)

        # Find and copy POM file
        pom_file_path = _find_pom_file(project_path)
        if not pom_file_path.exists():
            raise FileNotFoundError(f"pom.xml not found at {pom_file_path}")
        destination_pom = project_test_setup_path / 'pom.xml'
        shutil.copy(pom_file_path, destination_pom)

        logger.info(f"Created Maven structure at: {project_test_setup_path}")
        logger.info(f"Copied POM file from: {pom_file_path}")

        # Extract package name from source directory
        source_dir = Path(project_path)
        java_files = list(source_dir.rglob('*.java'))
        if not java_files:
            logger.warning("No Java source files found for extracting package name.")
            package_name = 'com.example'  # Default package name
        else:
            package_name = extract_package_name(java_files)
            if not package_name:
                package_name = 'com.example'  # Default package name

        # Adjust the pom.xml file with the new groupId
        success, error = adjust_pom_file(destination_pom, new_group_id=package_name)
        if not success:
            raise Exception(error)

        return True, ""
    except Exception as e:
        logger.error(f"Failed to create Maven structure: {e}")
        return False, str(e)


def copy_java_files(source_dir: Path, target_dir: Path) -> Tuple[int, bool, str]:
    """Copies all .java files from source to target directory."""
    try:
        java_files = list(source_dir.rglob('*.java'))
        count = 0
        for file in java_files:
            relative_path = file.relative_to(source_dir)
            destination = target_dir / relative_path
            destination.parent.mkdir(parents=True, exist_ok=True)
            shutil.copy(file, destination)
            count += 1
        logger.info(f"Copied {count} Java files from {source_dir} to {target_dir}")
        return count, True, ""
    except Exception as e:
        logger.error(f"Failed to copy Java files: {e}")
        return 0, False, str(e)

def extract_test_methods(test_code: str) -> List[Dict]:
    """Extracts individual test methods from test code."""
    test_methods = []
    lines = test_code.split('\n')
    current_method = {}
    in_method = False

    for i, line in enumerate(lines, 1):
        if '@Test' in line:
            in_method = True
            current_method = {'start_line': i}
            continue
        if in_method and ('public ' in line and '(' in line):
            method_name = line.strip().split('(')[0].split()[-1]
            current_method['name'] = method_name
            current_method['start_line'] = i
        if in_method and '}' in line:
            current_method['end_line'] = i
            test_methods.append(current_method)
            in_method = False

    return test_methods

def validate_test_structure(test_methods: List[Dict], test_code: str) -> bool:
    """Validates the structure of extracted test methods."""
    # Example validation checks
    for method in test_methods:
        if not method.get('name'):
            logger.error("Test method without a name detected.")
            return False
    # Additional validations can be added here
    logger.debug("All test methods validated successfully.")
    return True

def organize_test_data(test_methods: List[Dict], source_class: str) -> Dict:
    """Organizes extracted test methods into a structured format."""
    organized_data = {
        'source_class': source_class,
        'test_methods': test_methods,
        'test_suite': f"{source_class}Test"
    }
    return organized_data

def process_test_files(storage_path: Path, src_test_java: Path, iteration) -> Tuple[int, bool, str]:
    """Processes all test JSON files and writes source and test code."""
    try:
        # First, verify the base path exists
        if not (storage_path / 'unit_tests').exists():
            logger.error(f"unit_tests directory not found in {storage_path}")
            return 0, False, f"unit_tests directory not found in {storage_path}"

        # Get all iteration directories
        test_dirs = list(storage_path.glob('unit_tests/*_iteration'))

        # Log what was found
        logger.debug(f"Found {len(test_dirs)} test iteration directories: {[d.name for d in test_dirs]}")

        logger.debug(f"Processing test files for iteration {iteration}")

        if not test_dirs:
            logger.warning(f"No test iteration directories found in {storage_path}/unit_tests/")
            return 0, False, "No test iteration directories found."

        test_dir = storage_path / 'unit_tests' / f'{iteration}_iteration'

        files_copied = 0

        json_files = list(test_dir.glob('*.json'))
        for idx, json_file in enumerate(json_files):
            with open(json_file, 'r') as f:
                data = json.load(f)
            test_code = data['code'].get('testCode')
            class_name = data['metadata']['class_name']
            if test_code:
                test_methods = extract_test_methods(test_code)
                if validate_test_structure(test_methods, test_code):
                    # Generate a unique test class name
                    unique_class_name = f"{class_name}Test_{idx}"
                    # Modify the class name in test_code
                    modified_test_code = modify_class_name(test_code, unique_class_name)
                    # Define the test file path
                    test_file = src_test_java / f"{unique_class_name}.java"
                    test_file.parent.mkdir(parents=True, exist_ok=True)
                    with open(test_file, 'w') as tf:
                        tf.write(modified_test_code)
                    files_copied += 1
                    logger.debug(f"Wrote test code to {test_file}")
                else:
                    logger.error(f"Validation failed for test file: {json_file}")
                    return files_copied, False, f"Validation failed for {json_file}"

        return files_copied, True, ""
    except Exception as e:
        logger.error(f"Failed to process test files: {e}")
        return 0, False, str(e)

def modify_class_name(test_code: str, new_class_name: str) -> str:
    """Modifies the class name in the test code to the new class name."""
    lines = test_code.split('\n')
    for i, line in enumerate(lines):
        if 'public class' in line:
            parts = line.split()
            for j, part in enumerate(parts):
                if part == 'class' and j+1 < len(parts):
                    parts[j+1] = new_class_name
                    lines[i] = ' '.join(parts)
                    return '\n'.join(lines)
    return test_code

def init_test_env(target_dir: Path, source_project_main: Path, storage_path: Path, iteration_path: Path) -> bool:
    logger.info(f"Initializing test environment with:")
    logger.info(f"- Target directory: {target_dir}")
    logger.info(f"- Source project main: {source_project_main}")
    logger.info(f"- Storage path: {storage_path}")
    
    # Verify paths exist
    if not storage_path.exists():
        logger.error(f"Storage path does not exist: {storage_path}")
        return False

    if not (storage_path / 'unit_tests').exists():
        logger.error(f"unit_tests directory not found in storage path: {storage_path / 'unit_tests'}")
        return False
    project_test_setup_path = target_dir

    # Convert source_project_main to Path object if it's not already
    if not isinstance(source_project_main, Path):
        source_project_main = Path(source_project_main)

    # Step 1: Clean target directory
    success, error = clean_target_directory(target_dir)
    if not success:
        logger.error(f"Setup failed during cleanup: {error}")
        return False

    # Step 2: Create Maven structure
    success, error = create_maven_structure(project_test_setup_path, str(source_project_main))
    if not success:
        logger.error(f"Setup failed during Maven structure creation: {error}")
        return False

    # Step 3: Copy Java source files to Maven structure
    count, success, error = copy_java_files(source_project_main, project_test_setup_path / 'src' / 'main' / 'java')
    if not success:
        logger.error(f"Setup failed during copying Java files: {error}")
        return False

    # Step 4: Process test files
    src_test_java = project_test_setup_path / 'src' / 'test' / 'java'
    files_copied, success, error = process_test_files(storage_path, src_test_java, iteration_path)
    if not success:
        logger.error(f"Setup failed during processing test files: {error}")
        return False

    # Output summary
    logger.info(f"Setup Summary:")
    logger.info(f"Total Java files copied: {count + files_copied}")
    logger.info(f"Directory structure created at: {project_test_setup_path}")
    logger.info("Test environment setup completed successfully.")
    return True
