"""
Module for handling interactions with the OpenAI API.
Manages API requests and responses for test generation.
"""
import logging

from openai import OpenAI

import time

import torch
from transformers import AutoModelForCausalLM, AutoTokenizer, BitsAndBytesConfig, GenerationConfig

from LLMTests.utils.logging_utils import LoggerSetup
from LLMTests.config import Config

logger_setup = LoggerSetup(logger_name="LLMClient")
logger = logger_setup.get_logger()

class LLMClient:
    """
    Handles interactions with the OpenAI API.
    
    Attributes:
        client: OpenAI client instance
        api_params: Dictionary of API parameters
        model: OpenAI model identifier
    """
    
    def __init__(self, config: Config = None, long_config = None):
        """
        Initialize the LLM client with configuration.
        
        Args:
            config: Configuration object containing API settings
        """
        
        if config is None and not long_config is None:
            self.config = long_config
            if self.config['use_openai_api']:
                self.client = OpenAI(api_key=self.config['OPENAI_API_KEY'])
            if self.config['use_localai_api']:
                self.client = OpenAI(api_key=self.config['OPENAI_API_KEY'], base_url=self.config['localai_base_url'])
            if not self.config['use_openai_api'] and not self.config['use_localai_api']:
                self.client = None
                        
            if not self.config['use_openai_api'] and not self.config['use_localai_api']:
                self.model, self.tokenizer = self._load_model_and_tokenizer()
            else:
                self.model, self.tokenizer = None, None
        
        else:
            try:
                self.client = OpenAI(
                    api_key=config.OPENAI_API_KEY,
                    base_url=config.OPENAI_BASE_URL
                )
                
                self.api_params = {
                    "temperature": config.TEMPERATURE,
                    "max_tokens": config.MAX_TOKENS,
                    "top_p": config.TOP_P,
                    "frequency_penalty": config.FREQUENCY_PENALTY,
                    "presence_penalty": config.PRESENCE_PENALTY,
                    "stop": config.STOP
                }
                self.config['model_id'] = config.OPENAI_MODEL
                self.config['PAUSE_DURATION'] = 30
                self.config['NUM_RETRIES'] = 3
                logger.info("LLM client initialized successfully")
                
            except Exception as e:
                logger.error(f"Error initializing LLM client: {str(e)}")
                raise

    def generate_tests(self, prompt):
        """Generate test code using either OpenAI API or local model."""
        try:
            if self.config['use_openai_api'] or self.config['use_localai_api']:
                return self._generate_with_openai(prompt)
            else:
                return self._generate_from_local_model(prompt, self.config['model_temperature'])
        except Exception as e:
            logging.error(f"Error in generate_tests: {e}")
            return None

    def send_prompt(self, prompt: str):
        """
        Send prompt to OpenAI API and get response.
        
        Args:
            prompt: The prompt to send to the API
            
        Returns:
            Optional[str]: The API response content or None if error occurs
            
        Raises:
            Exception: If API call fails
        """
        return self.generate_tests(prompt)

####################################################################################################
    def _load_model_and_tokenizer(self):
        try:
            logging.info("Loading tokenizer...")
            cache_dir = self.config.get('cache_dir', './model_cache')
    
            # GPU availability check
            num_gpus = torch.cuda.device_count()
            logging.info(f"Number of GPUs available: {num_gpus}")
            
            # Configure device map based on GPU availability
            if (num_gpus > 1):
                device_map = "auto"
                logging.info("Using automatic device mapping for multiple GPUs")
            else:
                device_map = {"": 0} if torch.cuda.is_available() else "cpu"
                logging.info(f"Using device map: {device_map}")
    
            # Update config with actual device map
            self.config['device_map'] = device_map

            # Initialize tokenizer
            tokenizer = AutoTokenizer.from_pretrained(
                self.config['model_id'],
                cache_dir=cache_dir,
                trust_remote_code=True
            )

            bnb_config = BitsAndBytesConfig(
                load_in_4bit=self.config['use_4bit'],
                bnb_4bit_quant_type=self.config['bnb_4bit_quant_type'],
                bnb_4bit_compute_dtype=getattr(torch, self.config['compute_dtype']),
                bnb_4bit_use_double_quant=self.config['use_nested_quant'],
            )
            model = AutoModelForCausalLM.from_pretrained(
                self.config['model_id'],
                quantization_config=bnb_config,
                device_map=device_map,
                cache_dir=cache_dir,
                trust_remote_code=True,
                max_position_embeddings=8000  # Add maximum token length
            )
            model.config.use_cache = False
            return model, tokenizer
        except Exception as e:
            logging.error(f"Error loading model and tokenizer: {e}")
            raise    

    def _generate_with_openai(self, prompt, temperature=0.3):
        retry_count = 0
        while True:
            try:
                #logging.info(f"Prompt to send to OpenAI API: {prompt}")
                response = self.client.chat.completions.create(
                    model=self.config.get('model_id', 'gpt-4o-mini'),
                    messages=[
                        {"role": "system", "content": "You are an experienced software tester specializing in Android Applications."},
                        {"role": "user", "content": prompt}
                    ],
                    temperature=temperature
                )
                #logging.info(f"OpenAI API response: {response.choices[0].message.content}")
                return response.choices[0].message.content
            except Exception as e:
                logging.error(f"Error during OpenAI API call: {e}")
                retry_count += 1
                if retry_count >= self.config['NUM_RETRIES']:
                    raise
                logging.info(f"Retrying OpenAI API call... ({retry_count}/{self.config['NUM_RETRIES']})")
                time.sleep(self.config['PAUSE_DURATION'])


    def _generate_from_local_model(self, prompt, temperature=0.7):
        try:
            # Validate device availability
            device = torch.device("cuda:0" if torch.cuda.is_available() else "cpu")
            
            # Use model's actual max length
            max_length = min(4096, self.model.config.max_position_embeddings)
            
            encoded_input = self.tokenizer(
                prompt,
                return_tensors='pt',
                truncation=True,
                max_length=max_length
            )
            
            # Move to appropriate device
            encoded_input = {k: v.to(device) for k, v in encoded_input.items()}
    
            generation_config = GenerationConfig(
                max_new_tokens=self.config.get('max_tokens', 4096),
                temperature=temperature,
                top_p=self.config.get('top_p', 0.95),
                top_k=50,
                repetition_penalty=1.2,
                do_sample=True,
                pad_token_id=self.tokenizer.pad_token_id,
                eos_token_id=self.tokenizer.eos_token_id,
            )
    
            with torch.inference_mode():
                output_sequence = self.model.generate(
                    **encoded_input,
                    generation_config=generation_config,
                )[0]
    
            # Clean up CUDA cache
            if device.type == 'cuda':
                torch.cuda.empty_cache()
    
            decoded_output = self.tokenizer.decode(output_sequence, skip_special_tokens=True)
            
            # Extract response with validation
            response = decoded_output[len(prompt):].strip()
            if not response:
                logging.info("Empty response generated")
                return ""                
            return response
    
        except Exception as e:
            logging.error(f"Generation error: {str(e)}")
            raise RuntimeError(f"Error during text generation: {str(e)}")
        finally:
            # Ensure cleanup
            if 'encoded_input' in locals():
                del encoded_input