import re
from typing import Optional, Tuple

from LLMTests.utils.logging_utils import LoggerSetup
from LLMTests.utils.code_validator import CodeValidator
from LLMTests.utils.code_analyzer import CodeAnalyzer
from LLMTests.utils.code_repairer import CodeRepairer

logger_setup = LoggerSetup(logger_name="TestCodeProcessor")
logger = logger_setup.get_logger()

class TestCodeProcessor:
    """
    Handles test code extraction and validation.
    """

    def __init__(self, tab_size: int = 4):
        """
        Initialize the test code processor.

        Args:
            tab_size: The number of spaces to use for a tab.
        """
        # More permissive patterns
        self.code_block_pattern = r"```(?:java)?\s*\n(.*?)\n\s*```"
        self.direct_code_pattern = (
            r"(?:^|\n)\s*(?:package\s+[\w.]+;\s*(?:\n|$))?"  # Optional package declaration
            r"(?:(?:import\s+[\w.*]+;\s*(?:\n|$))*)"  # Import statements
            r"(?:public\s+)?(?:class|interface|enum)\s+\w+"  # Class/interface/enum declaration
        )
        self.tab_size = tab_size
        self.validator = CodeValidator()
        self.analyzer = CodeAnalyzer()
        self.repairer = CodeRepairer()
        logger.info("Test code processor initialized")

    def preprocess_response(self, response: str) -> str:
        """
        Preprocesses the LLM response to clean up code block delimiters.
        """
        response = response.replace("'''", "```")
        response = response.replace("'``", "```")
        response = response.replace("`'", "```")
        return response

    def extract_test_code(self, response: str) -> Optional[str]:
        """
        Extracts test code from LLM response with improved pattern matching.

        Args:
            response: Raw response from LLM

        Returns:
            Optional[str]: Extracted test code or None if no valid code found
        """
        try:
            if response is None or len(response) < 10:
                return response
            response = self.preprocess_response(response)
            logger.debug(f"Attempting to extract test code from response of length: {len(response)}")

            # First try code block pattern
            code_block_matches = re.findall(self.code_block_pattern, response, re.DOTALL)
            if code_block_matches:
                extracted_code = max(code_block_matches, key=len).strip()
                # Check if extracted code is likely Java code before returning
                if self.validator.is_likely_java_code(extracted_code):
                    return extracted_code
                else:
                    logger.warning("Extracted code block does not appear to be valid Java code.")

            # Then check for direct Java code
            if re.search(self.direct_code_pattern, response, re.MULTILINE):
                # If it looks like Java code, return the whole response
                cleaned = re.sub(r'^[^a-zA-Z]*', '', response)  # Remove leading non-letter chars
                cleaned = re.sub(r'\n+$', '\n', cleaned.rstrip())  # Clean trailing newlines
                if self.validator.is_likely_java_code(cleaned):
                    return cleaned
                else:
                    logger.warning("Direct code pattern match does not appear to be valid Java code.")

            # Finally, check if it's likely Java code
            if self.validator.is_likely_java_code(response):
                cleaned = response.strip()
                if cleaned.startswith(('package', 'import', 'public', 'class')):
                    return cleaned

            logger.warning(f"No valid test code patterns found in LLM response: {response[:100]}")
            return None

        except Exception as e:
            logger.error(f"Error extracting test code: {str(e)}", exc_info=True)
            return None

    def clean_extracted_code(self, code: str) -> str:
        """
        Cleans the extracted code by removing unwanted characters, ensuring proper indentation, and conditionally removing trailing text.

        Args:
            code: The extracted code to clean.

        Returns:
            The cleaned code.
        """
        if len(code) < 10:
            return code
        code = code.replace("`", "")
        code = code.replace("\t", " " * self.tab_size)
        code = code.replace("\r", "")
        lines = code.splitlines()
        cleaned_lines = [line.rstrip() for line in lines]

        # Conditionally remove trailing text after the last closing brace
        code = "\n".join(cleaned_lines)
        try:
            # Check if the code is likely a complete class
            if code.count("{") == code.count("}"):
                # Find the positions of all class declarations
                class_positions = []
                for match in re.finditer(r"(?:public\s+)?(?:class|interface|enum)\s+\w+", code):
                    class_positions.append(match.start())

                # Find the position of the last closing brace for each class
                last_brace_positions = []
                for class_start in class_positions:
                    brace_level = 0
                    last_brace_pos = -1
                    for i, char in enumerate(code[class_start:]):
                        if char == '{':
                            brace_level += 1
                        elif char == '}':
                            brace_level -= 1
                            if brace_level == 0:
                                last_brace_pos = class_start + i
                                break
                    if last_brace_pos != -1:
                        last_brace_positions.append(last_brace_pos)

                # Determine the overall last closing brace position
                if last_brace_positions:
                    overall_last_brace_pos = max(last_brace_positions)
                    # Find the line number of the overall last closing brace
                    last_brace_line = code.count('\n', 0, overall_last_brace_pos + 1) + 1
                    cleaned_lines = cleaned_lines[:last_brace_line]
                else:
                    logger.warning("Could not find any closing braces for class declarations.")

        except Exception as e:
            logger.error(f"Error removing trailing text: {e}")

        return "\n".join(cleaned_lines)

    def validate_test_code(self, code: Optional[str], language: str) -> Tuple[Optional[str], Optional[str]]:
        """
        Validates and cleans the extracted test code with enhanced error checking.
        """
        if not code and len(code) < 10:
            logger.warning("Validation failed: No test code was provided.")
            return None, "No test code provided"

        try:
            logger.debug(f"Starting validation of {language} test code (length: {len(code)})")
            code = code.replace("`", "")
            code = self.clean_extracted_code(code)

            # Detect language
            detected_language = self.validator.detect_language(code)
            if detected_language != "java":
                error_message = (
                    f"Unsupported language detected: {detected_language}. Only Java code is supported."
                )
                logger.error(error_message)
                return code, error_message

            # Run validation checks
            errors, fixes = self.analyzer.analyze_with_ast(code)
            has_assertions, issues = self.validator.check_for_assertions(code)
            
            if not has_assertions:
                issues.append("Include assertions using Assert.* methods to verify expected behavior")

            # Combine all issues
            all_issues = issues + errors
            if all_issues:
                logger.info(f"Found {len(all_issues)} issues to repair")
                code = self.repairer.auto_repair_code(code, {"issues": all_issues, **fixes})

                # Recheck after repair
                has_assertions, remaining_issues = self.validator.check_for_assertions(code)
                if remaining_issues:
                    error_message = "Test code validation failed:\n- " + "\n- ".join(remaining_issues)
                    logger.warning(error_message)
                    return code, error_message

            return code, None

        except Exception as e:
            error_message = f"Error validating {language} test code: {str(e)}."
            logger.error(error_message, exc_info=True)
            return None, error_message