# File: test_storage_manager.py
"""
Module for managing storage and organization of generated test files.
Handles saving test files, database interactions, and cleanup of old test iterations.
"""

import os
import json

from datetime import datetime
from typing import List, Tuple, Dict, Any
from dataclasses import dataclass

from LLMTests.utils.logging_utils import LoggerSetup
from LLMTests.config import Config
from LLMTests.extractMethods.method_data import MethodData  # Assuming we have a MethodData class for type hints

logger_setup = LoggerSetup(logger_name="TestStorageManager")
logger = logger_setup.get_logger()

@dataclass
class TestMetadata:
    """Data class for test metadata."""
    status: str = "pending"
    type: str = ""
    error_msg: str = ""
    version: str = "1.0"
    class_name: str = ""
    package: str = ""

class TestStorageManager:
    """
    Handles saving and managing test files.
    
    Attributes:
        storage_path: Base path for storing test files
        max_iterations: Maximum number of test iterations to keep
    """
    
    def __init__(self, storage_path: str, max_iterations: int = None):
        """
        Initialize the test storage manager.
        
        Args:
            storage_path: Base path for storing test files
            max_iterations: Maximum number of test iterations to keep (defaults to Config.MAX_ITERATIONS)
        """
        self.storage_path = storage_path
        self.max_iterations = max_iterations or Config.MAX_ITERATIONS
        self.test_base_path = os.path.join(storage_path, 'unit_tests')
        logger.info(f"Test storage manager initialized with storage path {storage_path}")

    def get_next_iteration(self) -> int:
        """
        Get the next iteration number based on existing directories.
        
        Returns:
            int: Next iteration number
        """
        iterations = [
            int(d.split('_')[0]) for d in os.listdir(self.test_base_path)
            if os.path.isdir(os.path.join(self.test_base_path, d)) and d.split('_')[0].isdigit()
        ]
        return max(iterations, default=0) + 1

    def create_test_content(self, language: str, unit_test: str, 
                          method_data: MethodData) -> Dict[str, Any]:
        """
        Create the content dictionary for a test file.
        
        Args:
            language: Programming language of the test
            unit_test: Generated test code
            method_data: Method information
            
        Returns:
            Dict[str, Any]: Formatted test content
        """
        metadata = TestMetadata(
            type=language,
            class_name=method_data.get_class_name(),
            package=method_data.get_package()
        )
        
        return {
            "metadata": {
                "status": metadata.status,
                "type": metadata.type,
                "error_msg": metadata.error_msg,
                "version": metadata.version,
                "class_name": metadata.class_name,
                "package": metadata.package
            },
            "code": {
                "testCode": unit_test,
                "sourceCode": method_data.get_source_code(),
                "generated_at": datetime.now().isoformat()
            }
        }

    def save_unit_tests(self, unit_tests: List[Tuple[str, str, MethodData]]) -> int:
        """
        Save generated unit tests to JSON files.
        
        Args:
            unit_tests: List of tuples containing (language, test_code, method_data)
            
        Returns:
            int: Current iteration number
            
        Raises:
            IOError: If file operations fail
        """
        try:
            # Ensure storage directory exists
            os.makedirs(self.test_base_path, exist_ok=True)
            
            # Get next iteration number
            iteration_count = self.get_next_iteration()
            iteration_folder = os.path.join(self.test_base_path, f"{iteration_count}_iteration")
            os.makedirs(iteration_folder, exist_ok=True)

            logger.info(f"Saving {len(unit_tests)} unit tests to iteration {iteration_count}")

            for idx, (language, unit_test, method_data) in enumerate(unit_tests, start=1):
                if not unit_test:
                    logger.warning(f"Skipping empty test at index {idx}")
                    continue

                try:
                    # Create test file
                    test_filename = os.path.join(iteration_folder, f"{idx}_test.json")
                    test_content = self.create_test_content(language, unit_test, method_data)

                    with open(test_filename, 'w') as test_file:
                        json.dump(test_content, test_file, indent=2)

                    logger.debug(f"Saved test file: {test_filename}")

                except IOError as e:
                    logger.error(f"Error saving test file {idx}: {str(e)}")
                    raise

            logger.info(f"Successfully saved all unit tests for iteration {iteration_count}")
            return iteration_count

        except Exception as e:
            logger.error(f"Error saving unit tests: {str(e)}")
            raise