# File: unit_test_generator.py
"""
Main module for coordinating the unit test generation process.
Orchestrates the interaction between various components to generate unit tests.
"""

import os
from typing import List, Optional, Tuple, Dict, Any
from dataclasses import dataclass

import re

from LLMTests.utils.logging_utils import LoggerSetup
from LLMTests.config import Config
from LLMTests.extractMethods.method_extractor import MethodExtractor
from LLMTests.extractMethods.prompt_generator import PromptGenerator
from LLMTests.utils.progress_tracker import ProgressTracker
from .llm_client import LLMClient
from .test_code_processor import TestCodeProcessor
from .test_storage_manager import TestStorageManager
from LLMTests.extractMethods.method_data import MethodData  # Assuming this exists from the original codebase
from LLMTests.extractMethods.test_combinator import combine_unit_tests

from LLMTests.testExecution.test_executor import TestExecutor

logger_setup = LoggerSetup(logger_name="UnitTestGenerator")
logger = logger_setup.get_logger()

@dataclass
class TestGenerationConfig:
    """Configuration settings for test generation."""
    n_shots: int = 0
    use_cot: bool = False  # Chain of Thought
    use_tas: bool = False  # Take a Step Back
    use_source_code: bool = True
    max_retries: int = 3

class UnitTestGenerator:
    """
    Main class coordinating the test generation process.
    
    This class orchestrates the interaction between various components:
    - Progress tracking
    - LLM interaction
    - Test code processing
    - Test storage management
    """
    
    def __init__(self, config: Config, storage_path: str, generation_config: Optional[TestGenerationConfig] = None, is_Demonstrator=True, long_config=None):
        """
        Initialize the unit test generator with all necessary components.
        
        Args:
            config: Main configuration object
            storage_path: Path to store generated tests
            generation_config: Optional test generation specific configuration
        """
        self.is_Demonstrator = is_Demonstrator
        self.config = config
        
        self.gen_config = generation_config or TestGenerationConfig()
            
        if self.is_Demonstrator:
            self.storage_path = storage_path

            self.storage_manager = TestStorageManager(storage_path)
        else:
            self.gen_config.n_shots = long_config.get('n_shots', 0)
            self.gen_config.use_cot = long_config.get('use_cot', False)
            self.gen_config.use_tas = long_config.get('use_tas', False)  
            self.gen_config.use_source_code = long_config.get('use_source_code', True)
            self.max_retries = long_config.get('NUM_RETRIES', 3)

        self.test_processor = TestCodeProcessor()
        # Initialize components
        self.progress_tracker = ProgressTracker()
        
        if not long_config is None:
            self.llm_client = LLMClient(long_config=long_config)
        else:
            self.llm_client = LLMClient(config)

        
        # State tracking
        self.unit_tests: List[Tuple[str, str, MethodData]] = []
        self.prompt_generator = PromptGenerator()
        
        logger_setup.set_log_level('DEBUG')  # Set desired log level here
        logger.info("UnitTestGenerator initialized with components")

    # def _combine_unit_tests(self, unit_tests: List[Tuple[str, str, object]]) -> str:
    #     """
    #     Combines a list of individual unit test strings into a single Java test class.

    #     Args:
    #         unit_tests: A list of tuples, where each tuple contains:
    #                     - The language (e.g., "java")
    #                     - The unit test code as a string
    #                     - The MethodData object (not used in this function, but might be useful for future extensions)

    #     Returns:
    #         A string containing the combined Java test class code.
    #     """

    #     combined_test_class = ""
    #     imports = set()
    #     test_methods = []
    #     class_name = ""

    #     for language, test_code, method_data in unit_tests:
    #         if language != "java":
    #             continue  # Skip non-Java tests for now

    #         # Extract package declaration
    #         package_match = re.search(r"package\s+([\w.]+);", test_code)
    #         if package_match and not combined_test_class:
    #             combined_test_class += f"{package_match.group(0)}\n\n"

    #         # Extract imports
    #         import_matches = re.findall(r"import\s+([\w.*]+);", test_code)
    #         for import_statement in import_matches:
    #             imports.add(import_statement)

    #         # Extract class name (assuming all tests are for the same class)
    #         class_name_match = re.search(r"public\s+class\s+(\w+)", test_code)
    #         if class_name_match and not class_name:
    #             class_name = class_name_match.group(1)

    #         # Extract test methods
    #         test_method_matches = re.findall(
    #             r"(@Test[\s\S]*?public\s+void\s+\w+\(\)[\s\S]*?})", test_code
    #         )
    #         test_methods.extend(test_method_matches)

    #     # Add static JUnit Assert import
    #     imports.add("static org.junit.Assert.*")

    #     # Add imports to the combined class
    #     for import_statement in sorted(imports):
    #         combined_test_class += f"import {import_statement};\n"

    #     # Add class declaration
    #     if class_name:
    #         combined_test_class += f"\npublic class {class_name} {{\n\n"
    #     else:
    #         combined_test_class += "\npublic class CombinedUnitTests {\n\n"

    #     # Add test methods
    #     for method in test_methods:
    #         combined_test_class += f"    {method}\n\n"

    #     # Close class
    #     combined_test_class += "}"

    #     return combined_test_class

    def remove_before_package(self, source_code):
        package_index = source_code.find("package")
        if package_index != -1:
            return source_code[package_index:]
        return source_code   

    def set_progress_callback(self, callback):
        """Set the progress callback function."""
        if self.is_Demonstrator:
            self.progress_tracker.set_progress_callback(callback)

    def count_testable_methods(self, path: str) -> Tuple[int, int]:
        """
        Count the total number of files and testable methods in a directory.
        
        Args:
            path: Directory path to scan
            
        Returns:
            Tuple[int, int]: (total_files, total_methods)
        """
        total_files = 0
        total_methods = 0
        
        try:
            for root, _, files in os.walk(path):
                for file in files:
                    if file.endswith(tuple(self.config.SUPPORTED_EXTENSIONS)):
                        total_files += 1
                        file_path = os.path.join(root, file)
                        
                        try:
                            method_extractor = MethodExtractor()
                            # if file is java but does not contain "test" in its name
                            if file.endswith(".java") and "test" not in file:
                                method_extractor.extract_methods_java(file_path)
                            elif file.endswith(".kt"):
                                method_extractor.extract_methods_kotlin(file_path)
                                
                            public_methods = method_extractor.get_public_methods()
                            total_methods += len(public_methods)
                            
                        except Exception as e:
                            logger.error(f"Error counting methods in {file_path}: {str(e)}")
            if self.is_Demonstrator:
                logger.info(f"Found {total_files} files with {total_methods} testable methods")
            return total_files, total_methods
            
        except Exception as e:
            logger.error(f"Error scanning directory {path}: {str(e)}")
            raise

    def process_file(self, file_path: str, is_Demonstrator=True) -> None:
        """
        Process a single file for test generation.
        
        Args:
            file_path: Path to the file to process
        """
        try:
            method_extractor = MethodExtractor()
            file_extension = os.path.splitext(file_path)[1]
            file_name = os.path.basename(file_path)

            # Extract methods based on file type
            if file_extension == ".java":
                method_extractor.extract_methods_java(file_path)
            elif file_extension == ".kt":
                method_extractor.extract_methods_kotlin(file_path)
            else:
                raise ValueError(f"Unsupported file type: {file_extension}")

            # Get methods and related information
            public_methods = method_extractor.get_public_methods()
            private_methods = method_extractor.get_private_methods()
            constructors = method_extractor.get_constructors()

            if not public_methods and self.is_Demonstrator:
                if self.is_Demonstrator:
                    self.progress_tracker.emit_progress('warning', 
                        f'No public methods found in {file_name}')
                return

            # Process each public method
            for method in public_methods:
                try:
                    if self.is_Demonstrator:
                        self.progress_tracker.current_method += 1
                        method_name = method.get_name()
                        class_name = method.get_class_name()
                                     
                        # Calculate percentage for progress tracking. Divide current_method by max(total_methods, 1) to avoid division by zero
                        calc_percent = (self.progress_tracker.current_method - 1) / max(self.progress_tracker.total_methods, 1)
            
                        percentage = 2 if self.progress_tracker.current_method == 1 else calc_percent * 100
                    
                        self.progress_tracker.emit_progress(
                            'generating',
                            f'Generating test for `{class_name}.{method_name}`',
                            {                                 
                             'percent': percentage,
                                'method_details': {
                                    'class_name': class_name,
                                    'method_name': method_name,
                                    'file_name': file_name,
                                    'current_method': self.progress_tracker.current_method,

                                    'total_methods': self.progress_tracker.total_methods
                                },
                                'current_method_name': method_name  # Added this line
                            }
                        )
                    # print the full method information:
                    #logger.info(f"Method: {method.get_name()} in class {method.get_class_name()} with signature \n{method.get_signature()}\nMethod Comments: \n{method.get_comments()}\n\nMethod Body:\n{method.get_method_body()}")
                    logger.info(f"Generating test for `{method.get_class_name()}.{method.get_name()}`")
                    
                    source_code = self.remove_before_package(method.get_source_code())
                    
                    method.set_source_code(source_code)
                    
                    unit_test = self.generate_test_for_method(
                        file_extension, method, constructors, private_methods, source_code=source_code
                    )
                    
                    if unit_test:
                        self.unit_tests.append((
                            "java" if file_extension == ".java" else "kotlin",
                            unit_test,
                            method
                        ))
                        
                        if self.is_Demonstrator:
                            # Report successful test generation
                            self.progress_tracker.emit_progress(
                                'generated',
                                f'Generated test for {class_name}.{method_name}',
                                {   'percent': int((self.progress_tracker.current_method - 2 / max(self.progress_tracker.total_methods, 1)) * 100),
                                    'method_details': {
                                        'class_name': class_name,
                                        'method_name': method_name,
                                        'status': 'completed',
                                        'current_method': self.progress_tracker.current_method,
                                        'total_methods': self.progress_tracker.total_methods,
                                    },
                                    'current_method_name': method_name  # Added this line
                                }
                            )
                        
                    else:
                        if self.is_Demonstrator:
                            # Report failed test generation
                            self.progress_tracker.emit_progress(
                                'warning',
                                f'Failed to generate test for {class_name}.{method_name}',
                                {
                                    'percent': int((self.progress_tracker.current_method - 2 / max(self.progress_tracker.total_methods, 1)) * 100),
                                    'method_details': {
                                        'class_name': class_name,
                                        'method_name': method_name,
                                        'status': 'failed',
                                        'current_method': self.progress_tracker.current_method,
                                        'total_methods': self.progress_tracker.total_methods,
                                    },
                                    'current_method_name': method_name  # Added this line
                                }
                            )

                except Exception as e:
                    logger.error(f"Error processing method in {file_name}: {str(e)}")
                    if self.is_Demonstrator:
                        self.progress_tracker.emit_progress(
                            'warning', 
                            f'Error processing method in {file_name}: {str(e)}',
                            {   
                            'percent': 0,
                                'method_details': {
                                    'class_name': method.get_class_name(),
                                    'method_name': method.get_name(),
                                    'status': 'error',
                                    'error': str(e)
                                },
                                'current_method_name': method.get_name()  # Added this line
                            }
                        )
            if not self.is_Demonstrator:
                return self.unit_tests

        except Exception as e:
            logger.error(f"Error processing file {file_path}: {str(e)}")
            raise

    def generate_test_for_method(
        self, file_extension: str, method: MethodData, 
        constructors: List[MethodData], private_methods: List[MethodData], source_code: str) -> Optional[str]:
        """
        Generate a test for a single method.
        
        Args:
            file_extension: File extension indicating language
            method: Method to generate test for
            constructors: Available constructors
            private_methods: Available private methods
            
        Returns:
            Optional[str]: Generated test code if successful, None otherwise
        """
        try:
            # Create prompt
            prompt = (self.prompt_generator.create_java_prompt if file_extension == ".java" 
                     else self.prompt_generator.create_kotlin_prompt)(
                method,
                constructors,
                private_methods,
                n_shots=self.gen_config.n_shots,
                use_cot=self.gen_config.use_cot,
                use_tas=self.gen_config.use_tas,
                use_source_code=self.gen_config.use_source_code
            )

            class_name = method.get_class_name()
            method_name = method.get_name()

            # Try generating test with retries
            for attempt in range(self.gen_config.max_retries):
                try:
                    #logger.debug(f"Attempt with prompt: {prompt}")
                    response = self.llm_client.send_prompt(prompt)
                    raw_test_code = self.test_processor.extract_test_code(response)
                    language = "java" if file_extension == ".java" else "kotlin"
                    
                    if raw_test_code:
                        validated_code, error_message = self.test_processor.validate_test_code(raw_test_code, language)
                        if not self.is_Demonstrator:
                            return validated_code
                    elif not self.is_Demonstrator and attempt < self.gen_config.max_retries:
                        attempt += 1
                        continue                        
                    test_executor = TestExecutor()
                    
                    # Run the test code
                    success, test_details, error = test_executor.run_single_test_code(
                        test_code=validated_code,
                        source_code=source_code  # The code under test
                    )
                    if success:
                        pass
                    else:
                        if error != "":
                            if self.is_Demonstrator:
                                # emit progress with error
                                self.progress_tracker.emit_progress(
                                    'warning',
                                    f'Error in test generation attempt {attempt + 1} for `{class_name}.{method_name}`: {error}',
                                    {
                                        'method_details': {
                                            'class_name': class_name,
                                            'method_name': method_name,
                                            'percent': int((self.progress_tracker.current_method - 2 / max(self.progress_tracker.total_methods, 1)) * 100),
                                            'attempt': attempt + 1,
                                            'max_attempts': self.gen_config.max_retries
                                        }
                                    }
                                )
                                
                            # the error should be added to the prompt and the prompt should be run again with a request to fix this error
                            prompt = f"""\n
                            You are an expert software tester specializing in java apps.\n
                            Fix the following JUnit4 test with the given error.\n
                            Return only the fully corrected code file without explanations.\n
                            Test Code:\n
                            {validated_code}\n\n
                            Error:\n
                            {error}                            
                            """
                            attempt = attempt=+1
                            continue
                            
                        else:
                            logger.error(f"Test failed successfully(?) for method `{class_name}.{method_name}` {error}")

                    
                    if validated_code:
                        if self.is_Demonstrator:
                            self.progress_tracker.emit_progress(
                            'generating',
                            f'Attempt {attempt + 1}/{self.gen_config.max_retries} for `{class_name}.{method_name}` succeeded',
                            {  
                                'method_details': {
                                    'class_name': class_name,
                                    'method_name': method_name,
                                    'percent': int((self.progress_tracker.current_method - 3 / max(self.progress_tracker.total_methods, 1)) * 100),
                                    'attempt': attempt + 1,
                                    'max_attempts': self.gen_config.max_retries
                                }
                            }
                        )
                        
                        return validated_code

                except Exception as e:
                    logger.error(
                        f"Error in test generation attempt {attempt + 1}: {str(e)}")

            return None

        except Exception as e:
            logger.error(f"Error generating test for method `{method.get_name()}`: {str(e)}")
            return None

    def process_directory(self, path: str) -> None:
        """
        Process all supported files in a directory for test generation.
        
        Args:
            path: Directory path to process
        """
        try:
            if not os.path.exists(path):
                raise FileNotFoundError(f"Directory not found: {path}")

            # Count total files and methods
            total_files, total_methods = self.count_testable_methods(path)
            
            if self.is_Demonstrator:
                self.progress_tracker.total_files = total_files
                self.progress_tracker.total_methods = total_methods

                if total_methods == 0:
                    self.progress_tracker.emit_progress('error', 'No testable methods found in project')
                    return

                self.progress_tracker.emit_progress(
                    'scanning',
                    f'Found {total_files} files with {total_methods} testable methods',
                    {
                        'percent': int(1)
                    }
                )

                # Process each file
                self.progress_tracker.current_file = 0
                self.progress_tracker.current_method = 0
            
            for root, _, files in os.walk(path):
                for file in files:
                    if file.endswith(tuple(self.config.SUPPORTED_EXTENSIONS)):
                        if self.is_Demonstrator:
                            self.progress_tracker.current_file += 1
                            self.progress_tracker.emit_progress(
                                'processing', 
                                f'Processing file {file} ({self.progress_tracker.current_file}/'
                                f'{total_files})',
                                {
                                    'percent': int(2)
                                }
                            )
                        
                        try:
                            self.process_file(os.path.join(root, file))
                        except Exception as e:
                            logger.error(f"Error processing file {file}: {str(e)}")
                            continue

            if self.is_Demonstrator:
                self.progress_tracker.emit_progress(
                    'complete', 
                    f'Completed processing {self.progress_tracker.current_method} methods from '
                    f'{self.progress_tracker.current_file} files',
                    {
                        'percent': int(100)
                    }
                )

        except Exception as e:
            logger.error(f"Error processing directory {path}: {str(e)}")
            raise

    def run(self, project_path: str) -> Dict[str, Any]:
        """
        Main method to run the unit test generation process.
        
        Args:
            project_path: Path to the project to process
            
        Returns:
            Dict[str, Any]: Contains status, iteration, and list of generated tests
        """
        try:
            logger.info(f"Starting unit test generation for project: {project_path}")
            
            # Process the project directory
            self.process_directory(project_path)
            
            # Save generated tests
            iteration = self.storage_manager.save_unit_tests(self.unit_tests)
            
            logger.info("Unit test generation completed successfully")
            return {
                'status': 'success',
                'iteration': iteration,
                'generated_tests': self.unit_tests  # List of (language, test_code, method_data)
            }
            
        except Exception as e:
            logger.error(f"Error in unit test generation process: {str(e)}")
            return {
                'status': 'error',
                'error': str(e)
            }