from typing import Dict, Tuple, List, Optional
import javalang
import re

from LLMTests.utils.logging_utils import LoggerSetup

logger_setup = LoggerSetup(logger_name="CodeAnalyzer")
logger = logger_setup.get_logger()

class CodeAnalyzer:
    def analyze_with_ast(self, code: str) -> Tuple[List[str], Dict[str, List[str]]]:
        """
        Analyzes the Java code using AST parsing to identify potential issues.

        Args:
            code: The Java code to analyze.

        Returns:
            A tuple containing:
            - A list of error messages.
            - A dictionary containing suggested fixes (e.g., missing imports, missing annotations).
        """
        errors = []
        fixes = {"add_annotations": [], "modify_methods": []}  # Removed "add_imports"
        try:
            tree = javalang.parse.parse(code)
            imported_packages = {imp.path for imp in tree.imports}
            imported_classes = {imp.path.split(".")[-1] for imp in tree.imports if not imp.path.endswith(".*")}

            for type_declaration in tree.types:
                if isinstance(type_declaration, javalang.tree.ClassDeclaration):
                    for member in type_declaration.body:
                        if isinstance(member, javalang.tree.MethodDeclaration):
                            # Check for missing @Test annotations
                            if member.name.startswith("test") and not any(
                                isinstance(ann, javalang.tree.Annotation) and ann.name == "Test"
                                for ann in member.annotations
                            ):
                                fixes["add_annotations"].append(
                                    f"Add @Test to method: {type_declaration.name}.{member.name}"
                                )

                            # Check for non-public test methods
                            if member.name.startswith("test") and "public" not in member.modifiers:
                                fixes["modify_methods"].append(
                                    f"Make method public: {type_declaration.name}.{member.name}"
                                )

                            # Check for missing imports and suggest them as comments
                            if member.body:
                                for statement in member.body:
                                    if isinstance(statement, javalang.tree.Statement):
                                        for path, node in statement.filter(javalang.tree.ReferenceType):
                                            type_name = node.name
                                            if type_name not in imported_packages and type_name not in imported_classes:
                                                # Check if the type is used in a way that requires an import
                                                if self.is_import_needed(statement, type_name):
                                                    errors.append(f"Potentially missing import for class: {type_name}")
                                                    # Suggest imports as comments
                                                    if "suggest_imports" not in fixes:
                                                        fixes["suggest_imports"] = []
                                                    fixes["suggest_imports"].extend(
                                                        [
                                                            f"// Consider adding import: {suggested_import}"
                                                            for suggested_import in self.suggest_import(
                                                                type_name, code
                                                            )
                                                        ]
                                                    )

        except javalang.parser.JavaSyntaxError as e:
            # More defensive error handling
            error_msg = str(e)
            if hasattr(e, 'description'):
                error_msg = e.description
            if hasattr(e, 'at') and e.at is not None and hasattr(e.at, 'line') and hasattr(e.at, 'column'):
                errors.append(f"Syntax error at line {e.at.line}, column {e.at.column}: {error_msg}")
            else:
                errors.append(f"Syntax error: {error_msg}")
        except Exception as e:
            errors.append(f"AST analysis error: {e}")

        return errors, fixes

    def analyze_missing_assertions(self, method_node: javalang.tree.MethodDeclaration) -> bool:
        """
        Analyzes a method AST node to detect if it's missing assertions.

        Args:
            method_node: javalang method declaration node

        Returns:
            bool: True if the method has no assertions
        """
        if not hasattr(method_node, "body") or not method_node.body:
            return True

        return not any(
            (
                node.qualifier in ["Assert", "Assertions"]
                or node.member.startswith("assert")
                or node.member
                in ["assertEquals", "assertTrue", "assertFalse", "assertNull", "assertNotNull", "assertThat"]
            )
            for path, node in method_node.filter(javalang.tree.MethodInvocation)
        )

    def suggest_assertion(self, method_node: javalang.tree.MethodDeclaration) -> Optional[str]:
        """
        Suggests an appropriate assertion based on method content and type inference.

        Args:
            method_node: javalang method declaration node

        Returns:
            Optional[str]: Suggested assertion statement
        """
        method_name = method_node.name.lower()

        # Analyze method body for clues
        if method_node.body:
            for statement in method_node.body:
                # Check for return statements
                if isinstance(statement, javalang.tree.ReturnStatement) and statement.expression:
                    try:
                        # Basic type inference
                        if isinstance(statement.expression, javalang.tree.Literal):
                            if statement.expression.value.lower() == "true":
                                return "Assert.assertTrue(result);"
                            elif statement.expression.value.lower() == "false":
                                return "Assert.assertFalse(result);"
                            elif statement.expression.value.isdigit():
                                return "Assert.assertEquals(expected, result);"  # Basic number check
                        elif isinstance(statement.expression, javalang.tree.MemberReference):
                            # If it's a simple variable, we might infer based on its name
                            if "null" in statement.expression.member.lower():
                                return "Assert.assertNull(result);"
                            elif any(
                                x in statement.expression.member.lower()
                                for x in ["true", "success", "valid"]
                            ):
                                return "Assert.assertTrue(result);"
                            elif any(x in statement.expression.member.lower() for x in ["false", "invalid"]):
                                return "Assert.assertFalse(result);"
                    except Exception as e:
                        logger.warning(f"Error during type inference: {e}")

        # Fallback to name-based heuristics
        if any(x in method_name for x in ["empty", "null"]):
            return "Assert.assertNull(result);"
        elif any(x in method_name for x in ["notnull", "exists", "present"]):
            return "Assert.assertNotNull(result);"
        elif any(x in method_name for x in ["true", "success"]):
            return "Assert.assertTrue(result);"
        elif any(x in method_name for x in ["false", "fail"]):
            return "Assert.assertFalse(result);"
        elif "equals" in method_name:
            return "Assert.assertEquals(expected, actual);"

        return "Assert.assertNotNull(result);"  # Generic suggestion

    def is_import_needed(self, statement: javalang.tree.Statement, type_name: str) -> bool:
        """
        Checks if a class usage within a statement strongly suggests that an import is needed.

        Args:
            statement: The AST statement node to analyze.
            type_name: The name of the class/type being used.

        Returns:
            True if an import is likely needed, False otherwise.
        """

        # Check for object instantiations
        for path, node in statement.filter(javalang.tree.ClassCreator):
            if node.type.name == type_name:
                return True

        # Check for static method calls or field accesses
        for path, node in statement.filter(javalang.tree.MemberReference):
            if node.qualifier == type_name:
                return True

        # Check for type declarations/references
        for path, node in statement.filter(javalang.tree.ReferenceType):
            if node.name == type_name:
                return True

        for path, node in statement.filter(javalang.tree.TypeArgument):
            if node.type and node.type.name == type_name:
                return True

        for path, node in statement.filter(javalang.tree.TypeParameter):
            if node.type and node.type.name == type_name:
                return True

        return False

    def suggest_import(self, class_name: str, code: str) -> List[str]:
        """
        Suggests potential imports for a given class name ONLY as comments.

        This version is EXTREMELY CONSERVATIVE and will only suggest imports as comments
        if it's highly likely they are needed, based on AST analysis.
        It also avoids duplicate suggestions.

        Args:
            class_name: The name of the class for which to suggest imports.
            code: The code being analyzed

        Returns:
            A list of potential import statements (as comments).
        """
        try:
            tree = javalang.parse.parse(code)
            imported_packages = {imp.path for imp in tree.imports}
            imported_classes = {imp.path.split(".")[-1] for imp in tree.imports if not imp.path.endswith(".*")}

            # Only suggest import if class is actually used in a way that requires it
            for type_declaration in tree.types:
                if isinstance(type_declaration, javalang.tree.ClassDeclaration):
                    for member in type_declaration.body:
                        if isinstance(member, javalang.tree.MethodDeclaration) and member.body:
                            for statement in member.body:
                                if isinstance(statement, javalang.tree.Statement):
                                    if self.is_import_needed(statement, class_name):
                                        # Prioritize existing packages
                                        suggestions = set()
                                        for existing_import in imported_packages:
                                            if "." in existing_import:
                                                package_prefix = existing_import.rsplit(".", 1)[0]
                                                suggestions.add(f"import {package_prefix}.{class_name};")

                                        # Convert to list to maintain order
                                        return list(suggestions)

        except Exception as e:
            logger.error(f"Error during AST analysis for import suggestion: {e}")

        return []  # Return empty list if no strong evidence found