import re
import javalang
from typing import Dict, List, Set

from LLMTests.utils.logging_utils import LoggerSetup
from LLMTests.utils.code_analyzer import CodeAnalyzer

logger_setup = LoggerSetup(logger_name="CodeRepairer")
logger = logger_setup.get_logger()

class CodeRepairer:
    def __init__(self):
        self.analyzer = CodeAnalyzer()

    def auto_repair_code(self, code: str, fixes: Dict[str, List[str]]) -> str:
        """
        Enhanced auto-repair function with NO automatic import adding, only suggestions as comments.
        """
        try:
            # First try to fix class structure if needed, but be more careful
            try:
                javalang.parse.parse(code)
            except javalang.parser.JavaSyntaxError as e:
                if "class, interface, or enum expected" in str(e):
                    code = self.fix_class_structure(code)

            lines = code.splitlines(True)
            changes_made = False

            # Get package name if present
            package_match = re.search(r"package\s+([\w.]+);", code)
            package_name = package_match.group(1) if package_match else None

            # Find existing imports safely
            existing_imports: Set[str] = set()
            import_matches = re.finditer(r"import\s+([\w.]+(?:\*)?);", code)
            existing_imports.update(match.group(0).strip() for match in import_matches if match)

            # Find import insertion point
            last_import_line = 0
            package_line = -1
            for i, line in enumerate(lines):
                if line.strip().startswith("package"):
                    package_line = i
                elif line.strip().startswith("import"):
                    last_import_line = i + 1
                elif line.strip() and not line.strip().startswith("package"):
                    break

            # If no imports found but we have a package, insert after package line
            if last_import_line == 0 and package_line != -1:
                last_import_line = package_line + 1

            # Process fixes (NO automatic import adding)
            if "suggest_imports" in fixes:
                added_suggestions = set()  # Track which suggestions we've already added
                for suggestion in fixes["suggest_imports"]:
                    # Skip if we've already added this suggestion or if it's already imported
                    suggestion_clean = suggestion.strip()
                    if (suggestion_clean in added_suggestions or 
                        any(suggestion_clean in imp for imp in existing_imports)):
                        continue
                    
                    # Only add if it's a valid import statement
                    if re.match(r'^import\s+[\w.]+(?:\*)?;$', suggestion_clean):
                        lines.insert(last_import_line, f"{suggestion_clean} // Suggested by CodeRepairer\n")
                        last_import_line += 1
                        added_suggestions.add(suggestion_clean)
                        changes_made = True

            # Process other fixes
            if "add_annotations" in fixes:
                for add_annotation in fixes["add_annotations"]:
                    if "Add @Test to method:" in add_annotation:
                        method_name = add_annotation.split("Add @Test to method: ")[1]

                        for i, line in enumerate(lines):
                            if method_name in line and "@Test" not in lines[i-1]:
                                lines.insert(i, "    @Test\n")
                                changes_made = True
                                break

            if "modify_methods" in fixes:
                for modify_method in fixes["modify_methods"]:
                    if "Make method public:" in modify_method:
                        method_name = modify_method.split("Make method public: ")[1]

                        for i, line in enumerate(lines):
                            if method_name in line and "public" not in line:
                                lines[i] = line.replace("private ", "public ")
                                lines[i] = line.replace("protected ", "public ")
                                changes_made = True
                                break

            if changes_made:
                logger.info("Auto-repair: Made modifications to fix issues")
                code = "".join(lines)
            else:
                logger.info("Auto-repair: No modifications needed")

            # Ensure package is on the first line only if needed
            code = self.move_package_declaration(code)
            return code

        except Exception as e:
            logger.error(f"Error in auto_repair_code: {str(e)}")
            return code

    def fix_class_structure(self, code: str) -> str:
        """
        Fixes missing class structure in Java code, but only if it's clearly needed.
        Wraps loose code in a proper class declaration if needed.
        Ensures the class name follows the <ClassName>Test convention.
        """
        try:
            # Skip if already has class structure or is likely a partial class
            if re.search(r'(?:public\s+)?class\s+\w+', code) or code.strip().endswith('{'):
                return code

            lines = code.splitlines()
            package_line = None
            import_lines = []
            code_lines = []
            has_loose_methods = False

            for i, line in enumerate(lines):
                if line.strip().startswith('package '):
                    package_line = line
                elif line.strip().startswith('import '):
                    import_lines.append(line)
                elif line.strip():
                    code_lines.append(line)
                    # Check for method definitions outside of a class
                    if re.match(r"^\s*(public|private|protected)?\s*(static)?\s*\w+\s+\w+\s*\(", line):
                        has_loose_methods = True

            # Only add class structure if we have loose methods
            if not has_loose_methods:
                return code

            # Build the proper class structure
            result = []
            if package_line:
                result.append(package_line)
            if import_lines:
                result.extend(import_lines)

            # Determine class name using multiple strategies
            class_name = self._determine_test_class_name(code_lines, package_line, import_lines)

            result.append("")  # Empty line before class
            result.append(f"public class {class_name} {{")

            # Indent and add the code
            indented_code = [f"    {line}" for line in code_lines]
            result.extend(indented_code)

            result.append("}")

            logger.info(f"Added class structure with name: {class_name}")
            return "\n".join(result)

        except Exception as e:
            logger.error(f"Error fixing class structure: {str(e)}")
            return code

    def _determine_test_class_name(self, code_lines: List[str], package_line: str = None, import_lines: List[str] = None) -> str:
        """Helper method to determine the appropriate test class name"""
        # Look for class references in code
        class_pattern = r'(?:new|instanceof)\s+(\w+)'
        method_pattern = r'@Test.*?void\s+test(\w+)'
        
        class_hints = []
        
        # Try to determine name from package first
        if package_line:
            package_parts = package_line.split()
            if len(package_parts) > 1:
                last_part = package_parts[-1].strip(';').split('.')[-1]
                if last_part:
                    class_hints.append(last_part)
        
        # If no package name found, look in code
        if not class_hints:
            for line in code_lines:
                class_matches = re.finditer(class_pattern, line)
                class_hints.extend(match.group(1) for match in class_matches)
                
                method_matches = re.finditer(method_pattern, line)
                class_hints.extend(match.group(1) for match in method_matches)
        
        # Select and clean up the name
        class_name = next((hint for hint in class_hints if hint), "MyClass")
        class_name = class_name.strip('Test')  # Remove Test if present
        class_name = class_name[0].upper() + class_name[1:] + "Test"  # Proper capitalization and add Test suffix
        
        return class_name

    def move_package_declaration(self, code: str) -> str:
        """
        Moves the package declaration to the first line of the file if it's not already there.

        Args:
            code (str): The content of the Java/Kotlin test file

        Returns:
            str: The modified content with package declaration as the first non-empty line
        """
        if not code:
            logger.warning("Cannot move package declaration: Empty code provided")
            return code

        try:
            logger.debug("Attempting to move package declaration")

            # Find package declaration using regex, handling potential comments on the same line
            package_match = re.search(r"^\s*package\s+([\w.]+);.*", code, re.MULTILINE)

            if not package_match:
                logger.debug("No package declaration found")
                return code

            package_line = package_match.group(0)
            package_name = package_match.group(1)

            # Check if already in the first non-empty line
            lines = code.splitlines(True)
            first_non_empty_line_index = -1
            for i, line in enumerate(lines):
                if line.strip():
                    first_non_empty_line_index = i
                    break

            if first_non_empty_line_index == -1:
                logger.debug("Empty code, adding package declaration")
                return f"package {package_name};\n"

            if lines[first_non_empty_line_index].strip().startswith(f"package {package_name}"):
                logger.debug("Package declaration already in the correct position")
                return code

            # Remove the package declaration from its current position
            code_without_package = code.replace(package_line, "", 1)

            # Insert package declaration at the beginning
            lines = code_without_package.splitlines(True)
            
            # Remove leading empty lines
            while lines and not lines[0].strip():
                lines.pop(0)

            lines.insert(0, f"package {package_name};\n")

            logger.info("Successfully moved package declaration to first non-empty line")
            return "".join(lines)

        except Exception as e:
            logger.error(f"Error moving package declaration: {str(e)}", exc_info=True)
            return code