import re
from typing import Dict, Tuple, List

from LLMTests.utils.logging_utils import LoggerSetup

logger_setup = LoggerSetup(logger_name="CodeValidator")
logger = logger_setup.get_logger()

class CodeValidator:
    def detect_language(self, code: str) -> str:
        """
        Detects the programming language of the given code snippet.

        Args:
            code: The code to analyze.

        Returns:
            The detected language ("java" or "unknown").
        """
        if any(
            marker in code for marker in ["public class", "import java", "@Test", "extends", "implements", ";"]
        ):
            return "java"
        else:
            return "unknown"

    def check_for_assertions(self, code: str) -> Tuple[bool, List[str]]:
        """
        Enhanced method to check for assertions and common test code issues.
        Improved to handle package-aware imports and common test patterns with robust error handling.
        """
        try:
            # List of common assertion patterns
            assertion_patterns = [
                r"Assert\.",  # Traditional Assert.* format
                r"assert[A-Z]\w+\(",  # Direct assertion calls like assertEquals(
                r"assert(?:Equals|True|False|Null|NotNull|That)\(",  # Common assertion methods
                r"@Test\(expected\s*=",  # Test expecting an exception
                r"fail\(",  # fail() method calls
                r"verify\(",  # Mockito verify
                r"should\.",  # BDD style assertions
                r"expect\(",  # Expectation style assertions
                r"assert\(",  # Plain assert statements
                r"assert\s*\(",  # Assert with whitespace
            ]

            issues = []

            # More comprehensive static import pattern matching
            static_import_patterns = [
                r"import\s+static\s+org\.junit\.Assert\.\*",
                r"import\s+static\s+junit\.framework\.Assert\.\*",
                r"import\s+static\s+org\.junit\.Assert\.[^;]+",
                r"import\s+static\s+junit\.framework\.Assert\.[^;]+",
            ]

            # Extract package name if present
            package_match = re.search(r"package\s+([\w.]+);", code)
            package_name = package_match.group(1) if package_match else None

            # Extract all imports safely
            import_matches = re.finditer(r"import\s+([\w.]+(?:\*)?);", code)
            imports = set(match.group(1) for match in import_matches if match)

            # Check for any form of static Assert import
            has_static_assert = any(re.search(pattern, code) for pattern in static_import_patterns)

            # Check for direct assertions
            direct_assertion_pattern = (
                r"(?m)^[^/]*(?:assertEquals|assertTrue|assertFalse|assertNull|assertNotNull|assertThat)\("
            )
            has_direct_assertions = bool(re.search(direct_assertion_pattern, code))

            # Suggest static import if needed
            if has_direct_assertions and not has_static_assert:
                issues.append("Consider adding 'import static org.junit.Assert.*' for cleaner assertion syntax")

            # Safely collect used classes
            used_classes = set()

            # Find 'new ClassName()' patterns
            new_class_matches = re.finditer(r"new\s+([A-Z][A-Za-z0-9_]*)", code)
            if new_class_matches:
                used_classes.update(match.group(1) for match in new_class_matches if match)

            # Find type declarations safely
            type_matches = re.finditer(r"(?:private|protected|public|)\s+([A-Z][A-Za-z0-9_]*)\s+\w+\s*[;=]", code)
            if type_matches:
                used_classes.update(match.group(1) for match in type_matches if match)

            # Define sets of classes to handle specially
            java_basic_types = {
                "String",
                "Integer",
                "Long",
                "Boolean",
                "Double",
                "Float",
                "Object",
                "Exception",
                "RuntimeException",
                "Throwable",
                "Error",
                "Class",
                "StringBuilder",
                "StringBuffer",
            }

            io_classes = {
                "Reader",
                "Writer",
                "InputStream",
                "OutputStream",
                "InputStreamReader",
                "OutputStreamWriter",
                "ByteArrayInputStream",
                "ByteArrayOutputStream",
                "BufferedReader",
                "BufferedWriter",
            }

            junit_classes = {
                "Test",
                "Before",
                "After",
                "BeforeClass",
                "AfterClass",
                "Rule",
                "ClassRule",
                "Assert",
                "Assume",
                "TestCase",
            }

            # Filter and check each used class
            filtered_classes = {cls for cls in used_classes if cls not in java_basic_types and cls.strip()}

            for class_name in filtered_classes:
                # Skip invalid class names
                if not class_name or not class_name[0].isupper() or "return" in class_name.lower():
                    continue

                try:
                    # Skip if class is in java.lang
                    if class_name in java_basic_types:
                        continue

                    # Skip if class is in the same package
                    if package_name and class_name:
                        package_class_pattern = rf"{re.escape(package_name)}\.{re.escape(class_name)}"
                        if re.search(package_class_pattern, code):
                            continue

                    # Skip if class is already imported
                    if any(imp.endswith(f".{class_name}") or imp.endswith(".*") for imp in imports):
                        continue

                    # Handle mock-related classes
                    if any(mock_word in class_name.lower() for mock_word in ["mock", "stub", "spy"]):
                        if not any("org.mockito" in imp for imp in imports):
                            if "Add Mockito imports for mock objects" not in issues:
                                issues.append("Add Mockito imports for mock objects")
                        continue

                    # Handle IO classes
                    if class_name in io_classes:
                        if not any("java.io" in imp for imp in imports):
                            issues.append(f"Add import java.io.{class_name}")
                        continue

                    # Handle JUnit classes
                    if class_name in junit_classes:
                        if not any("org.junit" in imp or "junit.framework" in imp for imp in imports):
                            issues.append(f"Add import org.junit.{class_name}")
                        continue

                    # For classes in the same package
                    if package_name and class_name:
                        # Only log as info if it might be in the same package
                        logger.debug(f"Class {class_name} might be in the same package ({package_name})")
                        continue

                    # Only add to issues if we're confident it's missing
                    if not any(imp.endswith(f".{class_name}") for imp in imports):
                        # Avoid duplicate issues
                        issue = f"Potentially missing import for class: {class_name}"
                        if issue not in issues:
                            issues.append(issue)

                except Exception as e:
                    logger.warning(f"Error processing class {class_name}: {str(e)}")
                    continue

            # Check for assertions
            try:
                code_without_comments = self.remove_comments_and_strings(code)
                has_assertions = any(re.search(pattern, code_without_comments) for pattern in assertion_patterns)
            except Exception as e:
                logger.warning(f"Error checking assertions: {str(e)}")
                has_assertions = False

            return has_assertions, issues

        except Exception as e:
            logger.error(f"Error in check_for_assertions: {str(e)}")
            return False, ["Error checking assertions and imports"]

    def check_test_structure(self, code: str) -> Dict[str, bool]:
        """
        Analyze the structure of a test class.

        Args:
            code: Test code to analyze

        Returns:
            Dict[str, bool]: Dictionary indicating presence of key test elements
        """
        logger.debug("Analyzing test class structure for completeness")

        structure = {
            "has_setup": "@Before" in code or "setUp()" in code or "@BeforeEach" in code,
            "has_teardown": "@After" in code or "tearDown()" in code or "@AfterEach" in code,
            "has_test_methods": "@Test" in code,
            "has_assertions": any(assert_keyword in code for assert_keyword in ["assert", "verify", "expect"]),
            "has_mocks": any(mock_keyword in code for mock_keyword in ["mock", "Mock", "stub", "spy"]),
        }

        logger.info(f"Test structure analysis complete: {structure}")
        if not structure["has_test_methods"]:
            logger.warning("No @Test methods found. Add at least one method annotated with @Test.")
        if not structure["has_assertions"]:
            logger.warning("No assertions found. Add assertions to verify expected behavior.")
        if not structure["has_setup"] and structure["has_test_methods"]:
            logger.info("No setup method found. Consider adding @Before if test initialization is needed.")
        if not structure["has_mocks"] and structure["has_test_methods"]:
            logger.info("No mocking detected. Consider using mocks for external dependencies.")

        return structure

    def remove_comments_and_strings(self, code: str) -> str:
        """
        Removes single-line comments, multi-line comments, and string literals from Java code.

        Args:
            code: The Java code to clean.

        Returns:
            The cleaned Java code with comments and strings removed.
        """
        # Remove single-line comments
        code = re.sub(r"//.*", "", code)

        # Remove multi-line comments
        code = re.sub(r"/\*.*?\*/", "", code, flags=re.DOTALL)

        # Remove string literals (handle escaped quotes)
        code = re.sub(r'"(?:\\.|[^"\\])*"', '""', code)

        return code

    def is_likely_java_code(self, code: str) -> bool:
        """
        Heuristically determines if a code snippet is likely Java code.

        Args:
            code: The code snippet to check.

        Returns:
            True if the code is likely Java, False otherwise.
        """
        java_keywords = ["public", "private", "protected", "class", "interface", "import", "package", "static", "void", "if", "else", "for", "while", "try", "catch", "finally", "return", "new", "this", "super"]
        return any(keyword in code for keyword in java_keywords)
