# logging_utils.py
import logging
import os
from datetime import datetime
from functools import wraps

class LoggerSetup:
    """Utility class for setting up logging across the application."""
    
    def __init__(self, log_dir=os.getcwd(), logger_name="LLLMTestModul"):
        """Initialize logger setup with directory for log files and logger name."""
        self.log_dir = log_dir
        #os.makedirs(log_dir, exist_ok=True)
        
        # Create timestamped log file
        timestamp = datetime.now().strftime("%Y%m%d_%H%M%S")
        #self.log_file = os.path.join(log_dir, f"{logger_name}_{timestamp}.log")
        
        # Configure logging
        self._configure_logger(logger_name)
        self._cleanup_old_logs()
        
    def _configure_logger(self, logger_name):
        """Configure the logger with file and console handlers."""
        self.logger = logging.getLogger(logger_name)
        self.logger.setLevel(logging.INFO)
        
        # Clear any existing handlers first
        self.logger.handlers.clear()
        
        # Console handler
        console_handler = logging.StreamHandler()
        console_handler.setFormatter(self._get_formatter())
        self.logger.addHandler(console_handler)
        
    @staticmethod
    def _get_formatter():
        """Get the log formatter with consistent format."""
        return logging.Formatter(
            '%(asctime)s - %(name)s - %(levelname)s - %(message)s'
        )
    
    def get_logger(self):
        """Get the configured logger instance."""
        return self.logger
    
    @staticmethod
    def log_method(logger):
        """Decorator for logging method entry, exit, and errors."""
        def decorator(func):
            @wraps(func)
            def wrapper(*args, **kwargs):
                method_name = func.__name__
                logger.debug(f"Entering method: {method_name}")
                try:
                    result = func(*args, **kwargs)
                    logger.debug(f"Exiting method: {method_name}")
                    return result
                except Exception as e:
                    logger.error(f"Error in method {method_name}: {str(e)}", exc_info=True)
                    raise
            return wrapper
        return decorator

    def _cleanup_old_logs(self, keep_last=5):
        """Remove old log files, keeping only the specified number of most recent logs."""
        try:
            log_files = [f for f in os.listdir(self.log_dir) if f.endswith('.log')]
            if len(log_files) <= keep_last:
                return
                
            log_files.sort(key=lambda x: os.path.getmtime(
                os.path.join(self.log_dir, x)
            ))
            
            for old_log in log_files[:-keep_last]:
                old_log_path = os.path.join(self.log_dir, old_log)
                os.remove(old_log_path)
                self.logger.debug(f"Removed old log file: {old_log}")
        except Exception as e:
            self.logger.error(f"Error cleaning up old logs: {str(e)}")

    def set_log_level(self, level: str):
        """Set the logging level dynamically."""
        try:
            level_upper = level.upper()
            if level_upper in ["DEBUG", "INFO", "WARNING", "ERROR", "CRITICAL"]:
                self.logger.setLevel(getattr(logging, level_upper))
                self.logger.info(f"Log level set to: {level_upper}")
            else:
                self.logger.warning(f"Invalid log level: {level}")
        except Exception as e:
            self.logger.error(f"Error setting log level: {str(e)}")